# -*- coding: utf-8 -*-
"""
TODO: box and whisker
http://tex.stackexchange.com/questions/115210/boxplot-in-latex
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from six.moves import range, map, zip
import os
import re
import textwrap
try:
    import numpy as np
except ImportError:
    pass
from os.path import join, splitext, dirname  # NOQA
from utool import util_num
from utool import util_inject
print, rrr, profile = util_inject.inject2(__name__)

#def ensure_latex_environ():
#    paths = os.environ['PATH'].split(os.pathsep)
#    mpl.rc('font',**{'family':'serif'})
#    mpl.rc('text', usetex=True)
#    mpl.rc('text.latex',unicode=True)
#    mpl.rc('text.latex',preamble='\usepackage[utf8]{inputenc}')


def find_ghostscript_exe():
    import utool as ut
    if ut.WIN32:
        gs_exe = r'C:\Program Files (x86)\gs\gs9.16\bin\gswin32c.exe'
    else:
        gs_exe = 'gs'
    return gs_exe


def compress_pdf(pdf_fpath, output_fname=None):
    """ uses ghostscript to write a pdf """
    import utool as ut
    ut.assertpath(pdf_fpath)
    suffix = '_' + ut.get_datestamp(False) + '_compressed'
    print('pdf_fpath = %r' % (pdf_fpath,))
    output_pdf_fpath = ut.augpath(pdf_fpath, suffix, newfname=output_fname)
    print('output_pdf_fpath = %r' % (output_pdf_fpath,))
    gs_exe = find_ghostscript_exe()
    cmd_list = (
        gs_exe,
        '-sDEVICE=pdfwrite',
        '-dCompatibilityLevel=1.4',
        '-dNOPAUSE',
        '-dQUIET',
        '-dBATCH',
        '-sOutputFile=' + output_pdf_fpath,
        pdf_fpath
    )
    ut.cmd(*cmd_list)
    return output_pdf_fpath


def make_full_document(text, title=None, preamp_decl={}, preamb_extra=None):
    r"""
    dummy preamble and document to wrap around latex fragment

    Args:
        text (str):
        title (str):

    Returns:
        str: text_

    CommandLine:
        python -m utool.util_latex --test-make_full_document

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_latex import *  # NOQA
        >>> text = 'foo'
        >>> title = 'title'
        >>> preamp_decl = {}
        >>> text_ = make_full_document(text, title)
        >>> result = str(text_)
        >>> print(result)
    """
    import utool as ut
    doc_preamb = ut.codeblock('''
    %\\documentclass{article}
    \\documentclass[10pt,twocolumn,letterpaper]{article}
    % \\usepackage[utf8]{inputenc}
    \\usepackage[T1]{fontenc}

    \\usepackage{times}
    \\usepackage{epsfig}
    \\usepackage{graphicx}
    \\usepackage{amsmath,amsthm,amssymb}
    \\usepackage[usenames,dvipsnames,svgnames,table]{xcolor}
    \\usepackage{multirow}
    \\usepackage{subcaption}
    \\usepackage{booktabs}

    %\\pagenumbering{gobble}
    ''')
    if preamb_extra is not None:
        if isinstance(preamb_extra, (list, tuple)):
            preamb_extra = '\n'.join(preamb_extra)
        doc_preamb += '\n' +  preamb_extra + '\n'
    if title is not None:
        preamp_decl['title'] = title

    decl_lines = [r'\{key}{{{val}}}'.format(key=key, val=val)
                  for key, val in preamp_decl.items()]
    doc_decllines = '\n'.join(decl_lines)

    doc_header = ut.codeblock(
        r'''
        \begin{document}
        ''')
    if preamp_decl.get('title') is not None:
        doc_header += r'\maketitle'

    doc_footer = ut.codeblock(
        r'''
        \end{document}
        ''')
    text_ = '\n'.join((doc_preamb, doc_decllines, doc_header, text, doc_footer))
    return text_


def render_latex_text(input_text, nest_in_doc=False, preamb_extra=None,
                      appname='utool', verbose=None):
    """ compiles latex and shows the result """
    import utool as ut
    if verbose is None:
        verbose = ut.VERBOSE
    dpath = ut.ensure_app_resource_dir(appname, 'latex_tmp')
    # put a latex framgent in a full document
    # print(input_text)
    fname = 'temp_render_latex'
    pdf_fpath = ut.compile_latex_text(
        input_text, dpath=dpath, fname=fname, preamb_extra=preamb_extra,
        verbose=verbose)
    ut.startfile(pdf_fpath)
    return pdf_fpath


def compile_latex_text(input_text, dpath=None, fname=None, verbose=True,
                       move=True, nest_in_doc=None, title=None,
                       preamb_extra=None):
    r"""
    CommandLine:
        python -m utool.util_latex --test-compile_latex_text --show

    Ignore:
        pdflatex -shell-escape --synctex=-1 -src-specials -interaction=nonstopmode\
            ~/code/ibeis/tmptex/latex_formatter_temp.tex

    Example1:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_latex import *  # NOQA
        >>> import utool as ut
        >>> verbose = True
        >>> #dpath = '/home/joncrall/code/ibeis/aidchallenge'
        >>> dpath = dirname(ut.grab_test_imgpath())
        >>> #ut.vd(dpath)
        >>> orig_fpaths = ut.list_images(dpath, fullpath=True)
        >>> figure_str = ut.get_latex_figure_str(orig_fpaths, width_str='2.4in', nCols=2)
        >>> input_text = figure_str
        >>> pdf_fpath = ut.compile_latex_text(input_text, dpath=dpath,
        >>>                                   verbose=verbose)
        >>> output_pdf_fpath = ut.compress_pdf(pdf_fpath)
        >>> print(pdf_fpath)
        >>> ut.quit_if_noshow()
        >>> ut.startfile(pdf_fpath)
    """
    import utool as ut
    if verbose:
        print('[ut] compile_latex_text')

    if nest_in_doc is None:
        nest_in_doc = 'documentclass' not in input_text
    if nest_in_doc:
        text = make_full_document(input_text, title=title,
                                  preamb_extra=preamb_extra)
    if not dpath:
        dpath = os.getcwd()
    if fname is None:
        fname = 'temp_latex'

    # Create temporary work directly
    work_dpath = join(dpath, '.tmptex')
    ut.ensuredir(work_dpath, verbose=verbose > 1)

    fname_tex = ut.ensure_ext(fname, '.tex')
    fname_pdf = ut.ensure_ext(fname, '.pdf')

    tex_fpath = join(work_dpath, fname_tex)
    pdf_fpath_output = join(work_dpath, fname_pdf)
    ut.write_to(tex_fpath, text)

    with ut.ChdirContext(work_dpath, verbose=verbose > 1):
        # print(text)
        args = ' '.join([
            'lualatex', '-shell-escape', '--synctex=-1', '-src-specials',
            '-interaction=nonstopmode', tex_fpath
        ])
        info = ut.cmd2(args, verbose=verbose > 1)
        if not ut.checkpath(pdf_fpath_output, verbose=verbose > 1):
            print('Error compiling LaTeX')
            ut.print_code(text, 'latex')
            print(info['out'])
            raise RuntimeError('latex failed ')

    if move:
        pdf_fpath = join(dpath, fname_pdf)
        ut.move(pdf_fpath_output, pdf_fpath, verbose=verbose > 1)
    else:
        pdf_fpath = pdf_fpath_output
    return pdf_fpath


def convert_pdf_to_image(pdf_fpath, ext='.jpg', verbose=1, dpi=300,
                         quality=90):
    import utool as ut
    if verbose:
        print('[ut] convert_pdf_to_image.')
    img_fpath = ut.ensure_ext(pdf_fpath, ext)
    if ut.UNIX:
        convert_fpath = ut.cmd2('which convert')['out'].strip()
        if not convert_fpath:
            raise Exception('ImageMagik convert was not found')
    args = ' '.join(['convert', '-density', str(dpi), pdf_fpath, '-quality',
                     str(quality), img_fpath])
    info = ut.cmd2(args, verbose=verbose > 1)  # NOQA
    if not ut.checkpath(img_fpath, verbose=verbose > 1):
        print('Failed to convert pdf to ' + ext)
        print(info['out'])
        raise Exception('ImageMagik failed to convert pdf to ' + ext)
    return img_fpath


def render_latex(input_text, dpath=None, fname=None, preamb_extra=None,
                 verbose=1, **kwargs):
    """
    Renders latex text into a jpeg.

    Whitespace that would have appeared in the PDF is removed, so the jpeg is
    cropped only the the relevant part.  This is ideal for figures that only
    take a single page.

    Args:
        input_text (?):
        dpath (str):  directory path(default = None)
        fname (str):  file name(default = None)
        preamb_extra (None): (default = None)
        verbose (int):  verbosity flag(default = 1)

    Returns:
        str: jpg_fpath -  file path string

    CommandLine:
        python -m utool.util_latex render_latex '$O(n^2)$' --fpath=~/slides/tmp.jpg

    Script:
        >>> # SCRIPT
        >>> from utool.util_latex import *  # NOQA
        >>> from os.path import split, expanduser
        >>> import utool as ut
        >>> input_text = ' '.join(ut.get_varargs()[1:])
        >>> dpath, fname = split(ut.argval('--fpath', ''))
        >>> dpath = expanduser(ut.argval('--dpath', dpath))
        >>> fname = ut.argval('--fname', fname)
        >>> kwargs = ut.dict_subset(ut.argparse_funckw(ut.convert_pdf_to_image), ['dpi', 'quality'])
        >>> jpg_fpath = render_latex(input_text, dpath, fname, **kwargs)
        >>> if ut.argflag('--diskshow'):
        >>>     ut.startfile(jpg_fpath)
    """
    import utool as ut
    try:
        import vtool_ibeis as vt
    except ImportError:
        import vtool as vt
    # turn off page numbers
    input_text_ = '\\pagenumbering{gobble}\n' + input_text
    # fname, _ = splitext(fname)
    img_fname = ut.ensure_ext(fname, ['.jpg'] + list(ut.IMG_EXTENSIONS))
    img_fpath = join(dpath, img_fname)
    pdf_fpath = ut.compile_latex_text(
        input_text_, fname=fname, dpath=dpath, preamb_extra=preamb_extra,
        verbose=verbose, move=False)
    ext = splitext(img_fname)[1]
    fpath_in = ut.convert_pdf_to_image(pdf_fpath, ext=ext, verbose=verbose)
    # Clip of boundaries of the pdf imag
    vt.clipwhite_ondisk(fpath_in, fpath_out=img_fpath, verbose=verbose > 1)
    return img_fpath


def latex_multicolumn(data, ncol=2, alignstr='|c|'):
    data = escape_latex(data)
    return r'\multicolumn{%d}{%s}{%s}' % (ncol, alignstr, data)


def latex_multirow(data, nrow=2):
    return r'\multirow{%d}{*}{%s}' % (nrow, data)


def latex_get_stats(lbl, data, mode=0):
    import utool as ut
    stats_ = ut.get_stats(data)
    if stats_.get('empty_list', False):
        return '% NA: latex_get_stats, data=[]'
    try:
        max_ = stats_['max']
        min_ = stats_['min']
        mean = stats_['mean']
        std  = stats_['std']
        shape = stats_['shape']
    except KeyError as ex:
        stat_keys = stats_.keys()  # NOQA
        ut.printex(ex, key_list=['stat_keys', 'stats_', 'data'])
        raise

    #int_fmt = lambda num: util.num_fmt(int(num))
    def float_fmt(num):
        return util_num.num_fmt(float(num))

    def tup_fmt(tup):
        return str(tup)
    fmttup = (float_fmt(min_), float_fmt(max_), float_fmt(mean), float_fmt(std), tup_fmt(shape))
    lll = ' ' * len(lbl)
    if mode == 0:
        prefmtstr = r'''
        {label} stats & min ; max = %s ; %s\\
        {space}       & mean; std = %s ; %s\\
        {space}       & shape = %s \\'''
    if mode == 1:
        prefmtstr = r'''
        {label} stats & min  = $%s$\\
        {space}       & max  = $%s$\\
        {space}       & mean = $%s$\\
        {space}       & std  = $%s$\\
        {space}       & shape = $%s$\\'''
    fmtstr = prefmtstr.format(label=lbl, space=lll)
    latex_str = textwrap.dedent(fmtstr % fmttup).strip('\n') + '\n'
    return latex_str


def latex_scalar(lbl, data):
    return (r'%s & %s\\' % (lbl, util_num.num_fmt(data))) + '\n'


def make_stats_tabular():
    'tabular for dipslaying statistics'
    pass


def ensure_rowvec(arr):
    arr = np.array(arr)
    arr.shape = (1, arr.size)
    return arr


def ensure_colvec(arr):
    arr = np.array(arr)
    arr.shape = (arr.size, 1)
    return arr


def escape_latex(text):
    r"""
    Args:
        text (str): a plain text message

    Returns:
        str: the message escaped to appear correctly in LaTeX

    References:
        http://stackoverflow.com/questions/16259923/how-can-i-escape-characters
    """
    conv = {
        '&': r'\&',
        '%': r'\%',
        '$': r'\$',
        '#': r'\#',
        '_': r'\_',
        '{': r'\{',
        '}': r'\}',
        '~': r'\textasciitilde{}',
        '^': r'\^{}',
        '\\': r'\textbackslash{}',
        '<': r'\textless',
        '>': r'\textgreater',
    }
    import six
    regex = re.compile('|'.join(re.escape(six.text_type(key)) for key in sorted(conv.keys(), key=lambda item: - len(item))))
    return regex.sub(lambda match: conv[match.group()], text)


def replace_all(str_, repltups):
    ret = str_
    for ser, rep in repltups:
        ret = re.sub(ser, rep, ret)
    return ret


def make_score_tabular(
        row_lbls, col_lbls, values, title=None, out_of=None, bold_best=False,
        flip=False, bigger_is_better=True, multicol_lbls=None, FORCE_INT=False,
        precision=None, SHORTEN_ROW_LBLS=False, col_align='l', col_sep='|',
        multicol_sep='|', centerline=True, astable=False, table_position='',
        AUTOFIX_LATEX=True, **kwargs):
    r"""
    makes a LaTeX tabular for displaying scores or errors

    Args:
        row_lbls (list of str):
        col_lbls (list of str):
        values (ndarray):
        title (str):  (default = None)
        out_of (None): (default = None)
        bold_best (bool): (default = True)
        flip (bool): (default = False)
        table_position (str) : eg '[h]'

    Returns:
        str: tabular_str

    CommandLine:
        python -m utool.util_latex --test-make_score_tabular:0 --show
        python -m utool.util_latex --test-make_score_tabular:1 --show
        python -m utool.util_latex --test-make_score_tabular:2 --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_latex import *  # NOQA
        >>> import utool as ut
        >>> row_lbls = ['config1', 'config2']
        >>> col_lbls = [r'score \leq 1', 'metric2']
        >>> values = np.array([[1.2, 2], [3.2, 4]])
        >>> title = 'title'
        >>> out_of = 10
        >>> bold_best = True
        >>> flip = False
        >>> tabular_str = make_score_tabular(row_lbls, col_lbls, values, title, out_of, bold_best, flip)
        >>> result = tabular_str
        >>> print(result)
        >>> ut.quit_if_noshow()
        >>> render_latex_text(tabular_str)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_latex import *  # NOQA
        >>> import utool as ut
        >>> row_lbls = ['config1']
        >>> col_lbls = [r'score \leq 1', 'metric2']
        >>> values = np.array([[1.2, 2]])
        >>> title = 'title'
        >>> out_of = 10
        >>> bold_best = True
        >>> flip = False
        >>> tabular_str = make_score_tabular(row_lbls, col_lbls, values, title, out_of, bold_best, flip)
        >>> result = tabular_str
        >>> print(result)
        >>> ut.quit_if_noshow()
        >>> render_latex_text(tabular_str)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_latex import *  # NOQA
        >>> import utool as ut
        >>> row_lbls = ['config1', 'config2']
        >>> col_lbls = [r'score \leq 1', 'metric2', 'foobar']
        >>> multicol_lbls = [('spam', 1), ('eggs', 2)]
        >>> values = np.array([[1.2, 2, -3], [3.2, 4, -2]])
        >>> title = 'title'
        >>> out_of = 10
        >>> bold_best = True
        >>> flip = False
        >>> tabular_str = make_score_tabular(row_lbls, col_lbls, values, title, out_of, bold_best, flip, multicol_lbls=multicol_lbls)
        >>> result = tabular_str
        >>> print(result)
        >>> ut.quit_if_noshow()
        >>> render_latex_text(tabular_str)
    """
    import utool as ut
    if flip:
        bigger_is_better = not bigger_is_better
        flip_repltups = [
            ('<=', '>'),
            ('>', '<='),
            ('\\leq', '\\gt'),
            ('\\geq', '\\lt'),
            ('score', 'error')
        ]
        col_lbls = [replace_all(lbl, flip_repltups) for lbl in col_lbls]
        if title is not None:
            title = replace_all(title, flip_repltups)
        if out_of is not None:
            values = out_of - values

    # Abbreviate based on common substrings
    common_rowlbl = None
    if SHORTEN_ROW_LBLS:
        if isinstance(row_lbls, list):
            row_lbl_list = row_lbls
        else:
            row_lbl_list = row_lbls.flatten().tolist()
        # Split the rob labels into the alg components
        #algcomp_list = [lbl.split(')_') for lbl in row_lbl_list]
        longest = long_substr(row_lbl_list)
        common_strs = []
        while len(longest) > 10:
            common_strs += [longest]
            row_lbl_list = [row.replace(longest, '...') for row in row_lbl_list]
            longest = long_substr(row_lbl_list)
        common_rowlbl = ('...'.join(common_strs)).replace(')_', ')_\n')
        row_lbls = row_lbl_list
        if len(row_lbl_list) == 1:
            common_rowlbl = row_lbl_list[0]
            row_lbls = ['0']

    # Stack values into a tabular body
    # TODO: need ability to specify datatypes
    def ensurelist(row_values):
        try:
            return row_values.tolist()
        except AttributeError:
            return row_values

    if False:
        # Numpy formatting
        def padvec(shape=(1, 1)):
            pad = np.array([[' ' for c in range(shape[1])] for r in range(shape[0])])
            return pad
        col_lbls = ensure_rowvec(col_lbls)
        row_lbls = ensure_colvec(row_lbls)
        _0 = np.vstack([padvec(), row_lbls])
        _1 = np.vstack([col_lbls, values])
        body = np.hstack([_0, _1])
        body = [[str_ for str_ in row] for row in body]
    else:
        assert len(row_lbls) == len(values)
        body = [[' '] + col_lbls]
        body += [[row_lbl] + ensurelist(row_values) for row_lbl, row_values in zip(row_lbls, values)]
    #import utool as ut
    # Fix things in each body cell
    DO_PERCENT = True
    try:
        for r in range(len(body)):
            for c in range(len(body[0])):
                # In data land
                if r > 0 and c > 0:
                    if precision is not None:
                        # Hack
                        if ut.is_float(body[r][c]):
                            fmtstr = '%.' + str(precision) + 'f'
                            body[r][c] = fmtstr % (float(body[r][c]),)
                    # Force integer
                    if FORCE_INT:
                        body[r][c] = str(int(float(body[r][c])))
                body[r][c] = str(body[r][c])
                # Remove bad formatting;
                if AUTOFIX_LATEX:
                    body[r][c] = escape_latex(body[r][c])
    except Exception as ex:
        import utool as ut
        print('len(row_lbls) = %r' % (len(row_lbls),))
        print('len(col_lbls) = %r' % (len(col_lbls),))
        print('len(values) = %r' % (values,))
        print('ut.depth_profile(values) = %r' % (ut.depth_profile(values),))
        ut.printex(ex, keys=['r', 'c'])
        raise

    # Bold the best values
    if bold_best:
        best_col_scores = values.max(0) if bigger_is_better else values.min(0)
        rows_to_bold = [np.where(values[:, colx] == best_col_scores[colx])[0]
                        for colx in range(len(values.T))]
        for colx, rowx_list in enumerate(rows_to_bold):
            for rowx in rowx_list:
                body[rowx + 1][colx + 1] = '\\txtbf{' + body[rowx + 1][colx + 1] + '}'

    # More fixing after the bold is in place
    for r in range(len(body)):
        for c in range(len(body[0])):
            # In data land
            if r > 0 and c > 0:
                if out_of is not None:
                    body[r][c] = body[r][c] + '/' + str(out_of)
                    if DO_PERCENT:
                        percent = ' = %.1f%%' % float(100 * values[r - 1, c - 1] / out_of)
                        body[r][c] += escape_latex(percent)

    # Align columns for pretty printing
    body = np.array(body)
    ALIGN_BODY = True
    if ALIGN_BODY:
        new_body_cols = []
        for col in body.T:
            colstrs = list(map(str, ensurelist(col)))
            collens = list(map(len, colstrs))
            maxlen = max(collens)
            newcols = [str_ + (' ' * (maxlen - len(str_))) for str_ in colstrs]
            new_body_cols += [newcols]
        body = np.array(new_body_cols).T

    # Build Body (and row layout)
    HLINE_SEP = True
    rowvalsep = ''
    colvalsep = ' & '
    endl = '\\\\\n'
    hline = r'\hline'
    #extra_rowsep_pos_list = [1]  # rows to insert an extra hline after
    extra_rowsep_pos_list = []  # rows to insert an extra hline after
    if HLINE_SEP:
        rowvalsep = hline + '\n'
    # rowstr list holds blocks of rows
    rowstr_list = [colvalsep.join(row) + endl for row in body]
    #rowstr_list = [row[0] + rowlbl_sep + colvalsep.join(row[1:]) + endl for row in body]
    #rowstr_list = [(
    #    ('' if len(row) == 0 else row[0])
    #    if len(row) <= 1 else
    #    row[0] + rowlblcol_sep + colvalsep.join(row[1:]) + endl)
    #    for row in body]
    rowsep_list = [rowvalsep for row in rowstr_list[0:-1]]  # should be len 1 less than rowstr_list
    # Insert multicolumn names
    if multicol_lbls is not None:
        # TODO: label of the row labels
        multicol_sep
        multicols = [latex_multicolumn(multicol, size, 'c' + multicol_sep) for multicol, size in multicol_lbls]
        multicol_str = latex_multirow('', 2) + colvalsep + colvalsep.join(multicols) + endl
        ncols = sum([tup[1] for tup in multicol_lbls])
        mcol_sep = '\\cline{2-%d}\n' % (ncols + 1,)
        rowstr_list = [multicol_str] + rowstr_list
        rowsep_list = [mcol_sep] + rowsep_list
        #extra_rowsep_pos_list += [1]

    # Insert title
    if title is not None and not astable:
        tex_title = latex_multicolumn(title, len(body[0])) + endl
        rowstr_list = [tex_title] + rowstr_list
        rowsep_list = [rowvalsep] + rowsep_list
        #extra_rowsep_pos_list += [2]

    # Apply an extra hline (for label)
    #extra_rowsep_pos_list = []
    for pos in sorted(extra_rowsep_pos_list)[::-1]:
        rowstr_list.insert(pos, '')
        rowsep_list.insert(pos, rowvalsep)
    #tabular_body = rowvalsep.join(rowstr_list)
    from six.moves import zip_longest
    tabular_body = ''.join([row if sep is None else row + sep for row, sep in zip_longest(rowstr_list, rowsep_list)])

    # Build Column Layout
    col_align_list = [col_align] * len(body[0])
    #extra_collayoutsep_pos_list = [1]
    extra_collayoutsep_pos_list = []
    for pos in  sorted(extra_collayoutsep_pos_list)[::-1]:
        col_align_list.insert(pos, '')
    #col_layaout_sep_list = rowlblcol_sep  # TODO

    rowlblcol_sep = '|'
    # Build build internal seprations between column alignments
    # Defaults to just the normal col_sep
    col_align_sep_list = [col_sep] * (len(col_align_list) - 1)
    # Adjust for the separations between row labels and the actual row data
    if len(col_align_sep_list) > 0:
        col_align_sep_list[0] = rowlblcol_sep
    # Continue multicolumn sepratation
    if multicol_lbls is not None:
        multicol_offsets = ut.cumsum(ut.get_list_column(multicol_lbls, 1))
        for offset in multicol_offsets:
            if offset < len(col_align_sep_list):
                col_align_sep_list[offset] = multicol_sep

    from six.moves import zip_longest
    _tmp = [ut.filter_Nones(tup) for tup in zip_longest(col_align_list, col_align_sep_list)]
    col_layout = ''.join(ut.flatten(_tmp))

    #if len(col_align_list) > 1:
    #    col_layout = col_align_list[0] + rowlblcol_sep + col_sep.join(col_align_list[1:])
    #else:
    #    col_layout = col_sep.join(col_align_list)

    tabular_head = (r'\begin{tabular}{|%s|}' % col_layout) + '\n'
    tabular_tail = r'\end{tabular}'

    if centerline:
        tabular_head = r'\centerline{' + '\n' + tabular_head
        tabular_tail = tabular_tail + '}'

    if astable:
        #tabular_head = r'\begin{centering}' + '\n' + tabular_head
        tabular_head = r'\centering' + '\n' + tabular_head
        tabular_head = r'\begin{table}' + table_position + '\n' + tabular_head

        lblstr = latex_sanitize_command_name(kwargs.get('label', title))
        caption = title
        if AUTOFIX_LATEX:
            caption = escape_latex(caption)
        caption = '\n% ---\n' + caption + '\n% ---\n'
        #tabular_head = r'\end{centering}' + '\n' + tabular_head
        tabular_tail = tabular_tail + '\n\\caption[%s]{%s}\n\\label{tbl:%s}\n\\end{table}' % (lblstr, caption, lblstr)

    tabular_str = rowvalsep.join([tabular_head, tabular_body, tabular_tail])
    topsep = '\\hline\n' if True else '\\toprule\n'
    botsep = '\\hline\n' if True else '\\bottomrule\n'
    tabular_str = tabular_head + topsep + tabular_body + botsep + tabular_tail

    if common_rowlbl is not None:
        #tabular_str += escape_latex('\n\nThe following parameters were held fixed:\n' + common_rowlbl)
        pass
    return tabular_str


def get_latex_figure_str2(fpath_list, cmdname, **kwargs):
    """ hack for candidacy """
    import utool as ut
    from os.path import relpath
    # Make relative paths
    if kwargs.pop('relpath', True):
        start = ut.truepath('~/latex/crall-candidacy-2015')
        fpath_list = [relpath(fpath, start) for fpath in fpath_list]
    cmdname = ut.latex_sanitize_command_name(cmdname)

    kwargs['caption_str'] = kwargs.get('caption_str', cmdname)
    figure_str  = ut.get_latex_figure_str(fpath_list, **kwargs)
    latex_block = ut.latex_newcommand(cmdname, figure_str)
    return latex_block


def get_latex_figure_str(fpath_list, caption_str=None, label_str=None,
                         width_str=r'\textwidth', height_str=None, nCols=None,
                         dpath=None, colpos_sep=' ', nlsep='',
                         use_sublbls=None, use_frame=False):
    r"""
    Args:
        fpath_list (list):
        dpath (str): directory relative to main tex file

    Returns:
        str: figure_str

    CommandLine:
        python -m utool.util_latex --test-get_latex_figure_str

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_latex import *  # NOQA
        >>> fpath_list = ['figures/foo.png']
        >>> figure_str = get_latex_figure_str(fpath_list)
        >>> result = str(figure_str)
        >>> print(result)
    """
    import utool as ut

    if nCols is None:
        nCols = len(fpath_list)

    USE_SUBFIGURE = True

    if width_str is not None:
        colwidth = (1.0 / nCols)
        if USE_SUBFIGURE:
            colwidth *= .95
            graphics_sizestr = ('%.2f' % (colwidth,)) + width_str
        else:
            graphics_sizestr = '[width=%.1f%s]' % (colwidth, width_str)
    elif height_str is not None:
        graphics_sizestr = '[height=%s]' % (height_str)
    else:
        graphics_sizestr =  ''

    if dpath is not None:
        fpath_list = [ut.relpath_unix(fpath_, dpath) for fpath_ in fpath_list]

    if USE_SUBFIGURE:
        # References: https://en.wikibooks.org/wiki/LaTeX/Floats,_Figures_and_Captions#Subfloats
        # TODO ? http://tex.stackexchange.com/questions/159290/how-can-i-place-a-vertical-rule-between-subfigures
        # Use subfigures
        graphics_list = []
        sublbl_prefix = label_str if label_str is not None else ''
        for count, fpath in enumerate(fpath_list):
            """
            print(', '.join([str(x) + ':' + chr(x) for x in range(65, 123)]))
            print(', '.join([str(x) + ':' + chr(x) for x in range(97, 123)]))
            """
            CHRLBLS = True
            if CHRLBLS:
                #subchar = chr(97 + count)
                subchar = chr(65 + count)
            else:
                subchar = str(count)
            parts = []
            subfigure_str = ''
            if len(fpath_list) > 1:
                parts.append('\\begin{subfigure}[h]{' + graphics_sizestr + '}')
                parts.append('\\centering')
            graphics_part = '\\includegraphics[width=%s]{%s}' % (width_str, fpath,)
            if use_frame:
                parts.append('\\fbox{%s}' % (graphics_part,))
            else:
                parts.append(graphics_part)
            if use_sublbls is True or use_sublbls is None and len(fpath_list) > 1:
                parts.append('\\caption{}\\label{sub:' + sublbl_prefix + subchar + '}')
            if len(fpath_list) > 1:
                parts.append('\\end{subfigure}')
            subfigure_str = ''.join(parts)
            graphics_list.append(subfigure_str)
    else:
        if True:
            graphics_list = [
                r'\includegraphics%s{%s}\captionof{figure}{%s}' % (
                    graphics_sizestr, fpath, 'fd',
                    #'(' + str(count) + ')'
                    #'(' + chr(97 + count) + ')'
                )
                for count, fpath in enumerate(fpath_list)]
        else:
            graphics_list = [r'\includegraphics%s{%s}' % (graphics_sizestr, fpath,) for fpath in fpath_list]
        #graphics_list = [r'\includegraphics%s{%s}' % (graphics_sizestr, fpath,) ]
    #nRows = len(graphics_list) // nCols

    # Add separators
    NL = '\n'
    if USE_SUBFIGURE:
        col_spacer_mid = NL + '~~' + '% --' + NL
        col_spacer_end = NL + r'\\' + '% --' + NL
    else:
        col_spacer_mid = NL + '&' + NL
        col_spacer_end = NL + r'\\' + nlsep + NL
    sep_list = [
        col_spacer_mid  if count % nCols > 0 else col_spacer_end
        for count in range(1, len(graphics_list) + 1)
    ]
    if len(sep_list) > 0:
        sep_list[-1] = ''
    graphics_list_ = [graphstr + sep for graphstr, sep in zip(graphics_list, sep_list)]

    #graphics_body = '\n&\n'.join(graphics_list)
    graphics_body = ''.join(graphics_list_)
    header_str = colpos_sep.join(['c'] * nCols)

    if USE_SUBFIGURE:
        figure_body = graphics_body
    else:
        figure_body =  ut.codeblock(
            r'''
            \begin{tabular}{%s}
            %s
            \end{tabular}
            '''
        ) % (header_str, graphics_body)
    if caption_str is not None:
        #tabular_body += '\n\caption{\\footnotesize{%s}}' % (caption_str,)
        if label_str is not None:
            figure_body += '\n\\caption[%s]{%s}' % (label_str, caption_str,)
        else:
            figure_body += '\n\\caption{%s}' % (caption_str,)
    if label_str is not None:
        figure_body += '\n\\label{fig:%s}' % (label_str,)
    #figure_fmtstr = ut.codeblock(
    #    r'''
    #    \begin{figure*}
    #    \begin{center}
    #    %s
    #    \end{center}
    #    \end{figure*}
    #    '''
    #)
    figure_fmtstr = ut.codeblock(
        r'''
        \begin{figure}[ht!]
        \centering
        %s
        \end{figure}
        '''
    )
    figure_str = figure_fmtstr % (figure_body)
    return figure_str


def long_substr(strlist):
    # Longest common substring
    # http://stackoverflow.com/questions/2892931/longest-common-substring-from-more-than-two-strings-python
    substr = ''
    if len(strlist) > 1 and len(strlist[0]) > 0:
        for i in range(len(strlist[0])):
            for j in range(len(strlist[0]) - i + 1):
                if j > len(substr) and is_substr(strlist[0][i:i + j], strlist):
                    substr = strlist[0][i:i + j]
    return substr


def is_substr(find, strlist):
    if len(strlist) < 1 and len(find) < 1:
        return False
    for i in range(len(strlist)):
        if find not in strlist[i]:
            return False
    return True


def tabular_join(tabular_body_list, nCols=2):
    dedent = textwrap.dedent
    tabular_head = dedent(r'''
    \begin{tabular}{|l|l|}
    ''')
    tabular_tail = dedent(r'''
    \end{tabular}
    ''')
    hline = ''.join([r'\hline', '\n'])
    tabular_body = hline.join(tabular_body_list)
    tabular = hline.join([tabular_head, tabular_body, tabular_tail])
    return tabular


def latex_newcommand(command_name, command_text, num_args=0):
    newcmd_str = '\\newcommand{\\' + command_name + '}'
    if num_args > 0:
        newcmd_str += '[%d]' % (num_args,)
    newcmd_str += '{'
    if '\n' in command_text:
        newcmd_str += '\n'
    newcmd_str += command_text
    if '\n' in command_text:
        newcmd_str += '\n'
    newcmd_str += '}'
    return newcmd_str


def latex_sanitize_command_name(_cmdname):
    r"""
    Args:
        _cmdname (?):

    Returns:
        ?: command_name

    CommandLine:
        python -m utool.util_latex --exec-latex_sanitize_command_name

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_latex import *  # NOQA
        >>> _cmdname = '#foo bar.'
        >>> command_name = latex_sanitize_command_name(_cmdname)
        >>> result = ('command_name = %s' % (str(command_name),))
        >>> print(result)
        FooBar
    """
    import utool as ut
    command_name = _cmdname
    try:
        def subroman(match):
            import roman
            try:
                groupdict = match.groupdict()
                num = int(groupdict['num'])
                if num == 0:
                    return ''
                return roman.toRoman(num)
            except Exception as ex:
                ut.printex(ex, keys=['groupdict'])
                raise
        command_name = re.sub(ut.named_field('num', r'\d+'), subroman, command_name)
    except ImportError as ex:
        if ut.SUPER_STRICT:
            ut.printex(ex)
            raise
    # remove numbers
    command_name = re.sub(r'[\d' + re.escape('#()[]{}.') + ']', '', command_name)
    # Remove _ for camel case
    #def to_camel_case(str_list):
    #    # hacky
    #    return ''.join([str_ if len(str_) < 1 else str_[0].upper() + str_[1:] for str_ in str_list])
    #command_name = to_cammel_case(re.split('[_ ]', command_name)[::2])
    str_list = re.split('[_ ]', command_name)
    #command_name = to_cammel_case(str_list)
    command_name = ut.to_camel_case('_'.join(str_list), mixed=True)
    return command_name


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_latex
        python -m utool.util_latex --allexamples
        python -m utool.util_latex --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
TODO: rectify name difference between parent and caller
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from six.moves import range, zip, map, filter  # NOQA
import fnmatch
import operator
import inspect
import traceback
import time
import sys
import six
import shelve
import textwrap
import keyword
import re
import types
import functools
try:
    import numpy as np
except ImportError:
    pass
from os.path import splitext, split, basename, dirname
from utool import util_inject
from utool import util_arg
from utool import util_list
from utool import util_print
from utool import util_str
from utool import util_type
from utool._internal import meta_util_six
print, rrr, profile = util_inject.inject2(__name__)

RAISE_ALL = util_arg.get_argflag('--raise-all', help='Causes ut.printex to always reraise errors')
FORCE_TB = util_arg.get_argflag('--force-tb', help='Causes ut.printex to always print traceback')
# COLORED_EXCEPTIONS = util_arg.get_argflag(('--colorex', '--cex'))
COLORED_EXCEPTIONS = not util_arg.get_argflag(('--no-colorex', '--no-cex'))


def print_traceback(with_colors=True):
    """
    prints current stack
    """
    #traceback.print_tb()
    import traceback
    stack = traceback.extract_stack()
    stack_lines = traceback.format_list(stack)
    tbtext = ''.join(stack_lines)
    if with_colors:
        try:
            from pygments import highlight
            from pygments.lexers import get_lexer_by_name
            from pygments.formatters import TerminalFormatter
            lexer = get_lexer_by_name('pytb', stripall=True)
            formatter = TerminalFormatter(bg='dark')
            formatted_text = highlight(tbtext, lexer, formatter)
            print(formatted_text)
        except Exception:
            print(tbtext)
    else:
        print(tbtext)


def ipython_execstr():
    return textwrap.dedent(r'''
    import sys
    embedded = False
    if '-w' in sys.argv or '--wait' in sys.argv or '--wshow' in sys.argv:
        print('waiting')
        in_ = raw_input('press enter')
    if '--cmd' in sys.argv or locals().get('in_', '') == 'cmd':
        print('[utool.dbg] Requested IPython shell with --cmd argument.')
        try:
            __IPYTHON__
            print('[ipython_execstr] Already in IPython!')
        except NameError:
            try:
                import IPython
                print('[utool.dbg] Presenting in new ipython shell.')
                embedded = True
                IPython.embed()
            except Exception as ex:
                print('[ipython_execstr]: Error: ' + six.text_type(type(ex)) + six.text_type(ex))
                raise
    ''')


def execstr_parent_locals():
    parent_locals = get_parent_frame().f_locals
    return execstr_dict(parent_locals, 'parent_locals')


def execstr_attr_list(obj_name, attr_list=None):
    execstr_list = [obj_name + '.' + attr for attr in attr_list]
    return execstr_list


varname_regex = re.compile('[_A-Za-z][_a-zA-Z0-9]*$')


def is_valid_varname(varname):
    """ Checks syntax and validity of a variable name """
    if not isinstance(varname, six.string_types):
        return False
    match_obj = re.match(varname_regex, varname)
    valid_syntax = match_obj is not None
    valid_name = not keyword.iskeyword(varname)
    isvalid = valid_syntax and valid_name
    return isvalid


def execstr_dict(dict_, local_name=None, exclude_list=None, explicit=False):
    """
    returns execable python code that declares variables using keys and values

    execstr_dict

    Args:
        dict_ (dict):
        local_name (str): optional: local name of dictionary. Specifying this
            is much safer
        exclude_list (list):

    Returns:
        str: execstr --- the executable string that will put keys from dict
            into local vars

    CommandLine:
        python -m utool.util_dbg --test-execstr_dict

    Example:
        >>> # DISABLE_DOCTEST
        >>> # UNSTABLE_DOCTEST
        >>> from utool.util_dbg import *  # NOQA
        >>> my_dictionary = {'a': True, 'b': False}
        >>> execstr = execstr_dict(my_dictionary)
        >>> exec(execstr)
        >>> assert 'a' in vars() and 'b' in vars(), 'execstr failed'
        >>> assert b is False and a is True, 'execstr failed'
        >>> result = execstr
        >>> print(result)
        a = my_dictionary['a']
        b = my_dictionary['b']

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_dbg import *  # NOQA
        >>> import utool as ut
        >>> my_dictionary = {'a': True, 'b': False}
        >>> execstr = execstr_dict(my_dictionary)
        >>> locals_ = locals()
        >>> exec(execstr, locals_)
        >>> a, b = ut.dict_take(locals_, ['a', 'b'])
        >>> assert 'a' in locals_ and 'b' in locals_, 'execstr failed'
        >>> assert b is False and a is True, 'execstr failed'
        >>> result = execstr
        >>> print(result)
        a = my_dictionary['a']
        b = my_dictionary['b']

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_dbg import *  # NOQA
        >>> import utool as ut
        >>> my_dictionary = {'a': True, 'b': False}
        >>> execstr = execstr_dict(my_dictionary, explicit=True)
        >>> result = execstr
        >>> print(result)
        a = True
        b = False
    """
    import utool as ut
    if explicit:
        expr_list = []
        for (key, val) in sorted(dict_.items()):
            assert isinstance(key, six.string_types), 'keys must be strings'
            expr_list.append('%s = %s' % (key, ut.repr2(val),))
        execstr = '\n'.join(expr_list)
        return execstr
    else:
        if local_name is None:
            # Magic way of getting the local name of dict_
            local_name = get_varname_from_locals(dict_, get_parent_frame().f_locals)
        try:
            if exclude_list is None:
                exclude_list = []
            assert isinstance(exclude_list, list)
            exclude_list.append(local_name)
            expr_list = []
            assert isinstance(dict_, dict), 'incorrect type type(dict_)=%r, dict_=%r' % (type(dict), dict_)
            for (key, val) in sorted(dict_.items()):
                assert isinstance(key, six.string_types), 'keys must be strings'
                if not is_valid_varname(key):
                    continue
                if not any((fnmatch.fnmatch(key, pat) for pat in exclude_list)):
                    expr = '%s = %s[%s]' % (key, local_name, ut.repr2(key))
                    expr_list.append(expr)
            execstr = '\n'.join(expr_list)
            return execstr
        except Exception as ex:
            locals_ = locals()
            ut.printex(ex, key_list=['locals_'])
            raise


def execstr_func(func):
    print(' ! Getting executable source for: ' + meta_util_six.get_funcname(func))
    _src = inspect.getsource(func)
    execstr = textwrap.dedent(_src[_src.find(':') + 1:])
    # Remove return statments
    while True:
        # Find first 'return'
        stmtx = execstr.find('return')
        if stmtx == -1:
            break  # Fail condition
        # The characters which might make a return not have its own line
        stmt_endx = len(execstr) - 1
        for stmt_break in '\n;':
            print(execstr)
            print('')
            print(stmtx)
            stmt_endx_new = execstr[stmtx:].find(stmt_break)
            if -1 < stmt_endx_new < stmt_endx:
                stmt_endx = stmt_endx_new
        # now have variables stmt_x, stmt_endx
        before = execstr[:stmtx]
        after  = execstr[stmt_endx:]
        execstr = before + after
    return execstr


def save_testdata(*args, **kwargs):
    """
    caches testdata
    """
    uid = kwargs.get('uid', '')
    shelf_fname = 'test_data_%s.shelf' % uid
    shelf = shelve.open(shelf_fname)
    locals_ = get_parent_frame().f_locals
    print('save_testdata(%r)' % (args,))
    for key in args:
        shelf[key] = locals_[key]
    shelf.close()


def load_testdata(*args, **kwargs):
    """
    tries to load previously cached testdata
    """
    uid = kwargs.get('uid', '')
    shelf_fname = 'test_data_%s.shelf' % uid
    shelf = shelve.open(shelf_fname)
    ret = [shelf[key] for key in args]
    shelf.close()
    if len(ret) == 1:
        ret = ret[0]
    print('load_testdata(%r)' % (args,))
    return ret


def import_testdata():
    shelf = shelve.open('test_data.shelf')
    print('importing\n * ' + '\n * '.join(shelf.keys()))
    shelf_exec = execstr_dict(shelf, 'shelf')
    exec(shelf_exec)
    shelf.close()
    return import_testdata.func_code.co_code


def breakpoint(*tags):
    import utool as ut
    if ut.get_argflag('--break'):
        ut.embed(N=1)
        return True
    return False


def fix_embed_globals():
    """
    HACK adds current locals() to globals().
    Can be dangerous.
    """
    frame = get_stack_frame(N=1)
    frame.f_globals.update(frame.f_locals)
    frame.f_globals['_did_embed_fix'] = True
    """
    def fix_embed_globals(N=0):
        import inspect
        # Get the parent frame
        frame_cur = inspect.currentframe()
        for _ix in range(N + 1):
            # always skip the frame of this function
            frame_next = frame_cur.f_back
            if frame_next is None:
                break
            frame_cur = frame_next
        # Add locals to parent globals
        frame = frame_cur
        frame.f_globals.update(frame.f_locals)
        frame.f_globals['_did_embed_fix'] = True
    """


def _wip_embed(parent_locals=None, parent_globals=None, exec_lines=None,
               remove_pyqt_hook=True, N=0):
    """
    Starts interactive session. Similar to keyboard command in matlab.
    Wrapper around IPython.embed

    Notes:
        #https://github.com/ipython/ipython/wiki/Cookbook%3a-Updating-code-for-use-with-IPython-0.11-and-later

        import IPython
        x = 3
        IPython.embed()
        c = IPython.Config()
        c.InteractiveShellApp.exec_lines = [
            '%pylab qt4',
            "print 'System Ready!'",
        ]

        def foo():
            return x + 3

        a = 3
        def bar():
            return a + 3
        bar()
        #NameError: global name 'a' is not defined


        from IPython.terminal.ipapp import TerminalIPythonApp
        x = 3
        app = TerminalIPythonApp.instance()
        app.initialize(argv=[]) # argv=[] instructs IPython to ignore sys.argv
        app.start()


    Args:
        parent_locals (None):
        parent_globals (None):
        exec_lines (None):
        remove_pyqt_hook (bool):
        N (int):

    CommandLine:
        python -m utool.util_dbg --test-embed

    References:
       http://stackoverflow.com/questions/27911570/can-you-specify-a-command-to-run-after-you-embed-into-ipython/27914204#27914204
       http://stackoverflow.com/questions/15167200/how-do-i-embed-an-ipython-interpreter-into-an-application-running-in-an-ipython

    Notes:
        Use cases I want to achieve

        1) Simply stop execution and embed in an IPython terminal session
        2) Like case 1, but execute a specific set of command (eg '%gui qt')
           AFTER IPython has started
        3) Embed and pause GUI execution (this is just case 1)
        3) Embed and let GUI execution continue while embeded. (basically just need case 2)

    TODO:
        try:
            get_ipython
        except NameError:
            banner=exit_msg=''
        else:
            banner = '*** Nested interpreter ***'
            exit_msg = '*** Back in main IPython ***'

        # First import the embed function
        from IPython.frontend.terminal.embed import InteractiveShellEmbed
        # Now create the IPython shell instance. Put ipshell() anywhere in your code
        # where you want it to open.
        ipshell = InteractiveShellEmbed(banner1=banner, exit_msg=exit_msg)
        #Then use ipshell() whenever you want to be dropped into an IPython shell. This
        #will allow you to embed (and even nest) IPython interpreters in your code and
        #inspect objects or the state of the program.

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_dbg import *  # NOQA
        >>> # build test data
        >>> parent_locals = None
        >>> parent_globals = None
        >>> exec_lines = None
        >>> remove_pyqt_hook = True
        >>> N = 0
        >>> # execute function
        >>> result = embed(parent_locals, parent_globals, exec_lines, remove_pyqt_hook, N)
        >>> # verify results
        >>> print(result)
    """
    import utool as ut
    from functools import partial
    import IPython

    if parent_globals is None:
        parent_globals = get_parent_frame(N=N).f_globals
    if parent_locals is None:
        parent_locals = get_parent_frame(N=N).f_locals

    stackdepth = N  # NOQA
    getframe = partial(ut.get_parent_frame, N=N)  # NOQA

    exec(execstr_dict(parent_globals, 'parent_globals'))
    exec(execstr_dict(parent_locals,  'parent_locals'))
    print('')
    print('================')
    print(ut.bubbletext('EMBEDDING'))
    print('================')
    print('[util] embedding')
    try:
        if remove_pyqt_hook:
            try:
                try:
                    import guitool_ibeis as gt
                except ImportError:
                    import guitool as gt
                gt.remove_pyqt_input_hook()
            except (ImportError, ValueError, AttributeError) as ex:
                #print(ex)
                printex(ex, iswarning=True)
                pass
            # make qt not loop forever (I had qflag loop forever with this off)
    except ImportError as ex:
        print(ex)
    user_ns = globals()
    user_ns = globals().copy()
    user_ns.update(locals())
    if parent_globals is not None:
        user_ns.update(parent_globals)
    if parent_locals is not None:
        user_ns.update(parent_locals)
    orig_argv = sys.argv  # NOQA
    print('About to start_ipython')
    config = IPython.Config()
    exec_lines_ = [
        '%pylab qt4',
        'print("Entered IPYTHON via utool")',
        'print("Entry Point: %r" % (ut.get_parent_frame(N=11).f_code.co_name,))',
        #'print("Entry Point: %r" % (ut.get_parent_frame(N=10).f_code.co_name,))',
        #'print("Entry Point: %r" % (ut.get_parent_frame(N=9).f_code.co_name,))',
        #'print("Entry Point: %r" % (ut.get_parent_frame(N=8).f_code.co_name,))',
        #'print("Entry Point: %r" % (ut.get_parent_frame(N=7).f_code.co_name,))',
        #'print("Entry Point: %r" % (ut.get_parent_frame(N=6).f_code.co_name,))',
        #'print("Entry Point: %r" % (ut.get_parent_frame(N=5).f_code.co_name,))',
        #execstr_dict(parent_locals)
    ] + ut.ensure_str_list(exec_lines if exec_lines is not None else [])
    config.InteractiveShellApp.exec_lines = exec_lines_
    print('Exec Lines: ')
    print(ut.indentjoin(exec_lines_, '\n    >>> '))
    IPython.start_ipython(config=config, argv=[], user_ns=user_ns)
    # Exit python immediately if specifed
    if user_ns.get('qqq', False) or vars.get('qqq', False) or user_ns.get('EXIT_NOW', False):
        print('[utool.embed] EXIT_NOW or qqq specified')
        sys.exit(1)


def embed(parent_locals=None, parent_globals=None, exec_lines=None,
          remove_pyqt_hook=True, N=0):
    """
    Starts interactive session. Similar to keyboard command in matlab.
    Wrapper around IPython.embed

    """
    import utool as ut
    from functools import partial
    import IPython

    if parent_globals is None:
        parent_globals = get_parent_frame(N=N).f_globals
    if parent_locals is None:
        parent_locals = get_parent_frame(N=N).f_locals

    stackdepth = N  # NOQA
    getframe = partial(ut.get_parent_frame, N=N)  # NOQA

    # exec(execstr_dict(parent_globals, 'parent_globals'))
    # exec(execstr_dict(parent_locals,  'parent_locals'))
    print('')
    print('================')
    print(ut.bubbletext('EMBEDDING'))
    print('================')
    print('[util] embedding')
    try:
        if remove_pyqt_hook:
            try:
                try:
                    import guitool_ibeis as gt
                except ImportError:
                    import guitool as gt
                gt.remove_pyqt_input_hook()
            except (ImportError, ValueError, AttributeError) as ex:
                #print(ex)
                printex(ex, iswarning=True)
                pass
            # make qt not loop forever (I had qflag loop forever with this off)
    except ImportError as ex:
        print(ex)

    #from IPython.config.loader import Config
    # cfg = Config()
    #config_dict = {}
    #if exec_lines is not None:
    #    config_dict['exec_lines'] = exec_lines
    #IPython.embed(**config_dict)
    print('[util]  Get stack location with: ')
    print('[util] ut.get_parent_frame(N=8).f_code.co_name')
    print('[util] set EXIT_NOW or qqq to True(ish) to hard exit on unembed')
    #print('set iup to True to draw plottool stuff')
    print('[util] call %pylab qt4 to get plottool stuff working')
    once = True
    # Allow user to set iup and redo the loop
    while once or vars().get('iup', False):
        if not once:
            # SUPER HACKY WAY OF GETTING FIGURES ON THE SCREEN BETWEEN UPDATES
            #vars()['iup'] = False
            # ALL YOU NEED TO DO IS %pylab qt4
            print('re-emebeding')
            for _ in range(100):
                time.sleep(.01)

        once = False
        #vars().get('iup', False):
        print('[util] calling IPython.embed()')
        """
        Notes:
            /usr/local/lib/python2.7/dist-packages/IPython/terminal/embed.py
            IPython.terminal.embed.InteractiveShellEmbed

            # instance comes from  IPython.config.configurable.SingletonConfigurable.instance
        """
        #c = IPython.Config()
        #c.InteractiveShellApp.exec_lines = [
        #    '%pylab qt4',
        #    '%gui qt4',
        #    "print 'System Ready!'",
        #]
        #IPython.embed(config=c)
        parent_ns = parent_globals.copy()
        parent_ns.update(parent_locals)
        locals().update(parent_ns)
        try:
            IPython.embed()
        except RuntimeError as ex:
            ut.printex(ex, 'Failed to open ipython')
        #config = IPython.terminal.ipapp.load_default_config()
        #config.InteractiveShellEmbed = config.TerminalInteractiveShell
        #module = sys.modules[parent_globals['__name__']]
        #config['module'] = module
        #config['module'] = module
        #embed2(stack_depth=N + 2 + 1)
        #IPython.embed(config=config)
        #IPython.embed(config=config)
        #IPython.embed(module=module)
        # Exit python immediately if specifed
        if vars().get('EXIT_NOW', False) or vars().get('qqq', False):
            print('[utool.embed] EXIT_NOW specified')
            sys.exit(1)


def embed2(**kwargs):
    """
    Modified from IPython.terminal.embed.embed so I can mess with stack_depth
    """
    config = kwargs.get('config')
    header = kwargs.pop('header', u'')
    stack_depth = kwargs.pop('stack_depth', 2)
    compile_flags = kwargs.pop('compile_flags', None)
    import IPython
    from IPython.core.interactiveshell import InteractiveShell
    from IPython.terminal.embed import InteractiveShellEmbed
    if config is None:
        config = IPython.terminal.ipapp.load_default_config()
        config.InteractiveShellEmbed = config.TerminalInteractiveShell
        kwargs['config'] = config
    #save ps1/ps2 if defined
    ps1 = None
    ps2 = None
    try:
        ps1 = sys.ps1
        ps2 = sys.ps2
    except AttributeError:
        pass
    #save previous instance
    saved_shell_instance = InteractiveShell._instance
    if saved_shell_instance is not None:
        cls = type(saved_shell_instance)
        cls.clear_instance()
    shell = InteractiveShellEmbed.instance(**kwargs)
    shell(header=header, stack_depth=stack_depth, compile_flags=compile_flags)
    InteractiveShellEmbed.clear_instance()
    #restore previous instance
    if saved_shell_instance is not None:
        cls = type(saved_shell_instance)
        cls.clear_instance()
        for subclass in cls._walk_mro():
            subclass._instance = saved_shell_instance
    if ps1 is not None:
        sys.ps1 = ps1
        sys.ps2 = ps2


def quitflag(num=None, embed_=False, parent_locals=None, parent_globals=None):
    if num is None or util_arg.get_argflag('--quit' + six.text_type(num)):
        if parent_locals is None:
            parent_locals = get_parent_frame().f_locals
        if parent_globals is None:
            parent_globals = get_parent_frame().f_globals
        exec(execstr_dict(parent_locals, 'parent_locals'))
        exec(execstr_dict(parent_globals, 'parent_globals'))
        if embed_:
            print('Triggered --quit' + six.text_type(num))
            embed(parent_locals=parent_locals,
                  parent_globals=parent_globals)
        print('Triggered --quit' + six.text_type(num))
        sys.exit(1)


def qflag(num=None, embed_=True):
    frame = get_parent_frame()
    return quitflag(num, embed_=embed_,
                    parent_locals=frame.f_locals,
                    parent_globals=frame.f_globals)


def quit(num=None, embed_=False):
    frame = get_parent_frame()
    return quitflag(num, embed_=embed_,
                    parent_locals=frame.f_globals,
                    parent_globals=frame.f_locals)


def in_jupyter_notebook():
    """
    http://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
    """
    try:
        cfg = get_ipython().config
        #print('cfg = %s' % (ut.repr4(cfg),))
        #x = cfg['IPKernelApp']['parent_appname']
        # might not work if using jupyter-console
        if cfg['IPKernelApp']['connection_file'].count('jupyter'):
            return True
        else:
            return False
    except (AttributeError, NameError):
        return False


def inIPython():
    """
    Tests if running in IPython
    """
    try:
        __IPYTHON__
        return True
    except NameError:
        return False


def haveIPython():
    """
    Tests if IPython is available
    """
    try:
        import IPython  # NOQA
        return True
    except NameError:
        return False


def print_frame(frame):
    frame = frame if 'frame' in vars() else inspect.currentframe()
    attr_list = ['f_code.co_name', 'f_code.co_filename', 'f_back', 'f_lineno',
                 'f_code.co_names']
    obj_name = 'frame'
    execstr_print_list = ['print("%r=%%r" %% (%s,))' % (_execstr, _execstr)
                          for _execstr in execstr_attr_list(obj_name, attr_list)]
    execstr = '\n'.join(execstr_print_list)
    exec(execstr)
    local_varnames = util_str.pack_into('; '.join(frame.f_locals.keys()))
    print('Local varnames: ' + local_varnames)
    print('--- End Frame ---')


def search_stack_for_localvar(varname):
    """
    Finds a local varable somewhere in the stack and returns the value

    Args:
        varname (str): variable name

    Returns:
        None if varname is not found else its value
    """
    curr_frame = inspect.currentframe()
    print(' * Searching parent frames for: ' + six.text_type(varname))
    frame_no = 0
    while curr_frame.f_back is not None:
        if varname in curr_frame.f_locals.keys():
            print(' * Found in frame: ' + six.text_type(frame_no))
            return curr_frame.f_locals[varname]
        frame_no += 1
        curr_frame = curr_frame.f_back
    print('... Found nothing in all ' + six.text_type(frame_no) + ' frames.')
    return None


def search_stack_for_var(varname, verbose=util_arg.NOT_QUIET):
    """
    Finds a varable (local or global) somewhere in the stack and returns the value

    Args:
        varname (str): variable name

    Returns:
        None if varname is not found else its value
    """
    curr_frame = inspect.currentframe()
    if verbose:
        print(' * Searching parent frames for: ' + six.text_type(varname))
    frame_no = 0
    while curr_frame.f_back is not None:
        if varname in curr_frame.f_locals.keys():
            if verbose:
                print(' * Found local in frame: ' + six.text_type(frame_no))
            return curr_frame.f_locals[varname]
        if varname in curr_frame.f_globals.keys():
            if verbose:
                print(' * Found global in frame: ' + six.text_type(frame_no))
            return curr_frame.f_globals[varname]
        frame_no += 1
        curr_frame = curr_frame.f_back
    if verbose:
        print('... Found nothing in all ' + six.text_type(frame_no) + ' frames.')
    return None

# Alias
get_var_from_stack = search_stack_for_var
get_localvar_from_stack = search_stack_for_localvar


def get_stack_frame(N=0, strict=True):
    """
    Args:
        N (int): N=0 means the frame you called this function in.
                 N=1 is the parent frame.
        strict (bool): (default = True)
    """
    frame_cur = inspect.currentframe()
    for _ix in range(N + 1):
        # always skip the frame of this function
        frame_next = frame_cur.f_back
        if frame_next is None:
            if strict:
                raise AssertionError('Frame level %r is root' % _ix)
            else:
                break
        frame_cur = frame_next
    return frame_cur


def get_parent_frame(N=0):
    parent_frame = get_stack_frame(N=N + 2)
    return parent_frame


def quasiquote(string):
    """ mimics lisp quasi quote functionality """
    return string.format(**get_parent_frame().f_locals)


fmtlocals = quasiquote  # non-lispy alias for quasiquote


def get_caller_prefix(N=0, aserror=False):
    prefix_fmt = '[!%s]' if aserror else '[%s]'
    prefix = prefix_fmt % (get_caller_name(N=N + 1),)
    return prefix


def get_caller_lineno(N=0, strict=True):
    parent_frame = get_stack_frame(N=N + 2)
    lineno =  parent_frame.f_lineno
    return lineno


def get_caller_name(N=0, allow_genexpr=True):
    """
    get the name of the function that called you

    Args:
        N (int): (defaults to 0) number of levels up in the stack
        allow_genexpr (bool): (default = True)

    Returns:
        str: a function name

    CommandLine:
        python -m utool.util_dbg get_caller_name
        python -m utool get_caller_name
        python ~/code/utool/utool/__main__.py get_caller_name
        python ~/code/utool/utool/__init__.py get_caller_name
        python ~/code/utool/utool/util_dbg.py get_caller_name

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_dbg import *  # NOQA
        >>> import utool as ut
        >>> N = list(range(0, 13))
        >>> allow_genexpr = True
        >>> caller_name = get_caller_name(N, allow_genexpr)
        >>> print(caller_name)
    """
    if isinstance(N, (list, tuple, range)):
        name_list = []
        for N_ in N:
            try:
                name_list.append(get_caller_name(N_))
            except AssertionError:
                name_list.append('X')
        return '[' + ']['.join(name_list) + ']'
    parent_frame = get_stack_frame(N=N + 2)
    caller_name = parent_frame.f_code.co_name
    co_filename = parent_frame.f_code.co_filename

    if not allow_genexpr:
        count = 0
        while True:
            count += 1
            if caller_name == '<genexpr>':
                parent_frame = get_stack_frame(N=N + 1 + count)
                caller_name = parent_frame.f_code.co_name
            else:
                break
    #try:
    #    if 'func' in  parent_frame.f_locals:
    #        caller_name += '(' + meta_util_six.get_funcname(parent_frame.f_locals['func']) + ')'
    #except Exception:
    #    pass
    if caller_name == '<module>':
        # Make the caller name the filename
        caller_name = splitext(split(co_filename)[1])[0]
    if caller_name in {'__init__', '__main__'}:
        # Make the caller name the filename
        caller_name = basename(dirname(co_filename)) + '.' + caller_name
    return caller_name


def get_caller_modname(N=0, allowmain=True):
    parent_frame = get_stack_frame(N=N + 2)
    assert allowmain is True
    caller_modname = parent_frame.f_globals['__name__']
    return caller_modname


def get_current_stack_depth():
    import traceback
    stack = traceback.extract_stack()
    return len(stack)


def module_functions(module):
    module_members = inspect.getmembers(module)
    function_list = []
    for key, val in module_members:
        if inspect.isfunction(val) and inspect.getmodule(val) == module:
            function_list.append((key, val))
    return function_list


def public_attributes(input):
    public_attr_list = []
    all_attr_list = dir(input)
    for attr in all_attr_list:
        if attr.find('__') == 0:
            continue
        public_attr_list.append(attr)
    return public_attr_list


def explore_stack():
    stack = inspect.stack()
    tup = stack[0]
    for ix, tup in reversed(list(enumerate(stack))):
        frame = tup[0]
        print('--- Frame %2d: ---' % (ix))
        print_frame(frame)
        print('\n')
        #next_frame = curr_frame.f_back


def explore_module(module_, seen=None, maxdepth=2, nonmodules=False):
    def __childiter(module):
        for aname in iter(dir(module)):
            if aname.find('_') == 0:
                continue
            try:
                yield module.__dict__[aname], aname
            except KeyError as ex:
                print(repr(ex))
                pass

    def __explore_module(module, indent, seen, depth, maxdepth, nonmodules):
        valid_children = []
        ret = u''
        modname = six.text_type(module.__name__)
        #modname = repr(module)
        for child, aname in __childiter(module):
            try:
                childtype = type(child)
                if not isinstance(childtype, types.ModuleType):
                    if nonmodules:
                        fullstr = indent + '    ' + six.text_type(aname) + ' = ' + repr(child)
                        truncstr = util_str.truncate_str(fullstr) + '\n'
                        ret +=  truncstr
                    continue
                childname = six.text_type(child.__name__)
                if seen is not None:
                    if childname in seen:
                        continue
                    elif maxdepth is None:
                        seen.add(childname)
                if childname.find('_') == 0:
                    continue
                valid_children.append(child)
            except Exception as ex:
                print(repr(ex))
                pass
        # Print
        ret += indent + modname + '\n'
        # Recurse
        if maxdepth is not None and depth >= maxdepth:
            return ret
        ret += ''.join([__explore_module(child,
                                         indent + '    ',
                                         seen, depth + 1,
                                         maxdepth,
                                         nonmodules)
                       for child in iter(valid_children)])
        return ret
    #ret +=
    #print('#module = ' + str(module_))
    ret = __explore_module(module_, '     ', seen, 0, maxdepth, nonmodules)
    #print(ret)
    sys.stdout.flush()
    return ret


def debug_npstack(stacktup):
    print('Debugging numpy [hv]stack:')
    print('len(stacktup) = %r' % len(stacktup))
    for count, item in enumerate(stacktup):
        if isinstance(item, np.ndarray):
            print(' * item[%d].shape = %r' % (count, item.shape))
        elif isinstance(item, list) or isinstance(item, tuple):
            print(' * len(item[%d]) = %d' % (count, len(item)))
            print(' * DEBUG LIST')
            with util_print.Indenter(' * '):
                debug_list(item)
        else:
            print(' *  type(item[%d]) = %r' % (count, type(item)))


def debug_list(list_):
    dbgmessage = []
    append = dbgmessage.append
    append('debug_list')
    dim2 = None
    if all([util_type.is_listlike(item) for item in list_]):
        append(' * list items are all listlike')
        all_lens = [len(item) for item in list_]
        if util_list.allsame(all_lens):
            dim2 = all_lens[0]
            append(' * uniform lens=%d' % dim2)
        else:
            append(' * nonuniform lens = %r' % np.unique(all_lens).tolist())
    else:
        all_types = [type(item) for item in list_]
        if util_list.allsame(all_types):
            append(' * uniform types=%r' % all_types[0])
        else:
            append(' * nonuniform types: %r' % np.unique(all_types).tolist())
    print('\n'.join(dbgmessage))
    return dim2


def debug_hstack(stacktup):
    try:
        return np.hstack(stacktup)
    except ValueError as ex:
        print('ValueError in debug_hstack: %s' % (ex,))
        debug_npstack(stacktup)
        raise


def debug_vstack(stacktup):
    try:
        return np.vstack(stacktup)
    except ValueError as ex:
        print('ValueError in debug_vstack: %s' % (ex,))
        debug_npstack(stacktup)
        raise


def debug_exception(func):
    @functools.wraps(func)
    def ex_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as ex:
            import utool
            msg = ('[tools] ERROR: %s(%r, %r)' % (meta_util_six.get_funcname(func), args, kwargs))
            #print(msg)
            utool.printex(ex, msg)
            #print('[tools] ERROR: %r' % ex)
            raise
    return ex_wrapper


TB = util_arg.get_flag('--tb')


def printex(ex, msg='[!?] Caught exception', prefix=None, key_list=[],
            locals_=None, iswarning=False, tb=TB, pad_stdout=True, N=0,
            use_stdout=False, reraise=False, msg_=None, keys=None,
            colored=None):
    """
    Prints (and/or logs) an exception with relevant info

    Args:
        ex (Exception): exception to print
        msg (str): a message to display to the user
        keys (None): a list of strings denoting variables or expressions of interest
        iswarning (bool): prints as a warning rather than an error if True (defaults to False)
        tb (bool): if True prints the traceback in the error message
        pad_stdout (bool): separate the error message from the rest of stdout with newlines
        prefix (None):
        locals_ (None):
        N (int):
        use_stdout (bool):
        reraise (bool):
        msg_ (None):
        key_list (list): DEPRICATED use keys

    Returns:
        None
    """
    import utool as ut
    if isinstance(ex, MemoryError):
        ut.print_resource_usage()
    if keys is not None:
        # shorthand for key_list
        key_list = keys
    # Get error prefix and local info
    if prefix is None:
        prefix = get_caller_prefix(aserror=True, N=N)
    if locals_ is None:
        locals_ = get_parent_frame(N=N).f_locals
    # build exception message
    if msg is True:
        key_list = get_parent_frame().f_locals
        msg = msg_
    exstr = formatex(ex, msg, prefix, key_list, locals_, iswarning, tb=tb, colored=colored)
    # get requested print function
    if use_stdout:
        def print_func(*args):
            msg = ', '.join(list(map(six.text_type, args)))
            sys.stdout.write(msg + '\n')
            sys.stdout.flush()
    else:
        print_func = ut.partial(ut.colorprint, color='yellow' if iswarning else 'red')
        # print_func = print
    if pad_stdout:
        print_func('\n+------\n')
    # print the execption
    print_func(exstr)
    if pad_stdout:
        print_func('\nL______\n')
    # If you dont know where an error is coming from raise-all
    if (reraise and not iswarning) or RAISE_ALL:
        sys.stdout.flush()
        sys.stderr.flush()
        raise ex
    if ut.get_argflag('--exit-on-error'):
        print('WARNING: dont use this flag. Some errors are meant to be caught')
        ut.print_traceback()
        print('REQUESTED EXIT ON ERROR')
        sys.exit(1)


def formatex(ex, msg='[!?] Caught exception',
             prefix=None, key_list=[], locals_=None, iswarning=False, tb=False,
             N=0, keys=None, colored=None):
    r"""
    Formats an exception with relevant info

    Args:
        ex (Exception): exception to print
        msg (unicode):  a message to display to the user (default = u'[!?] Caught exception')
        keys (None): a list of strings denoting variables or expressions of interest (default = [])
        iswarning (bool): prints as a warning rather than an error if True (default = False)
        tb (bool): if True prints the traceback in the error message (default = False)
        prefix (None): (default = None)
        locals_ (None): (default = None)
        N (int): (default = 0)
        colored (None): (default = None)
        key_list (list): DEPRICATED use keys

    Returns:
        str: formated exception

    CommandLine:
        python -m utool.util_dbg --exec-formatex

    Example:
        >>> # ENABLE_DOCTET
        >>> from utool.util_dbg import *  # NOQA
        >>> import utool as ut
        >>> msg = 'Testing Exception'
        >>> prefix = '[test]'
        >>> key_list = ['N', 'foo', 'tb']
        >>> locals_ = None
        >>> iswarning = True
        >>> keys = None
        >>> colored = None
        >>> def failfunc():
        >>>     tb = True
        >>>     N = 0
        >>>     try:
        >>>         raise Exception('test exception. This is not an error')
        >>>     except Exception as ex:
        >>>         result = formatex(ex, msg, prefix, key_list, locals_,
        >>>                           iswarning, tb, N, keys, colored)
        >>>         return result
        >>> result = failfunc().replace('\n\n', '')
        >>> print(result)

        <!!! WARNING !!!>
        Traceback (most recent call last):
          File "<string>", line 15, in failfunc
        Exception: test exception. This is not an error[test] Testing Exception
        <class 'Exception'>: test exception. This is not an error
        [test] N = 0
        [test] foo = NameError (this likely due to a misformatted printex and is not related to the exception)
        [test] tb = True
        </!!! WARNING !!!>

    """
    # Get error prefix and local info
    if prefix is None:
        prefix = get_caller_prefix(aserror=True, N=N)
    if locals_ is None:
        locals_ = get_parent_frame(N=N).f_locals
    if keys is not None:
        # shorthand for key_list
        key_list = keys
    # build exception message
    errstr_list = []  # list of exception strings
    ex_tag = 'WARNING' if iswarning else 'EXCEPTION'
    errstr_list.append('<!!! %s !!!>' % ex_tag)
    if tb or FORCE_TB:
        tbtext = traceback.format_exc()
        if colored or COLORED_EXCEPTIONS:
            from utool import util_str
            tbtext = util_str.highlight_text(tbtext, lexer_name='pytb', stripall=True)
        errstr_list.append(tbtext)
    errstr_list.append(prefix + ' ' + six.text_type(msg) + '\n%r: %s' % (type(ex), six.text_type(ex)))
    #errstr_list.append(prefix + ' ' + six.text_type(msg) + '\ntype(ex)=%r' % (type(ex),))
    parse_locals_keylist(locals_, key_list, errstr_list, prefix)
    errstr_list.append('</!!! %s !!!>' % ex_tag)
    return '\n'.join(errstr_list)


def get_varname_from_stack(var, N=0, **kwargs):
    return get_varname_from_locals(var, get_parent_frame(N=N).f_locals, **kwargs)


def get_varname_from_locals(val, locals_, default='varname-not-found',
                            strict=False, cmpfunc_=operator.is_):
    """ Finds the string name which has where locals_[name] is val

    Check the varname is in the parent namespace
    This will only work with objects not primatives

    Args:
        val (): some value
        locals_ (dict): local dictionary to search
        default (str):
        strict (bool):

    Returns:
        str: the varname which is Val (if it exists)

    """
    if val is None or isinstance(val, (int, float, bool)):
        # Cannot work on primative types
        return default
    try:
        for count, val_ in enumerate(six.itervalues(locals_)):
            if cmpfunc_(val, val_):
                index_ = count
        varname = six.text_type(list(locals_.keys())[index_])
    except NameError:
        varname = default
        if strict:
            raise
    return varname


def get_varval_from_locals(key, locals_, strict=False):
    """
    Returns a variable value from locals.
    Different from locals()['varname'] because
    get_varval_from_locals('varname.attribute', locals())
    is allowed
    """
    assert isinstance(key, six.string_types), 'must have parsed key into a string already'
    if key not in locals_:
        dotpos = key.find('.')
        if dotpos > -1:
            key_ = key[:dotpos]
            attrstr_ = key[dotpos:]
            try:
                baseval = locals_[key_]  # NOQA
                val = eval('baseval' + attrstr_)
            except Exception as ex:
                if strict:
                    raise
                val = ex
        else:
            raise AssertionError('%s = NameError' % (key))
    else:
        val = locals_[key]
    return val


def get_varstr(val, pad_stdout=True, locals_=None):
    # TODO: combine with printex functionality
    if locals_ is None:
        locals_ = get_parent_frame().f_locals
    name = get_varname_from_locals(val, locals_)
    varstr_list = []
    if pad_stdout:
        varstr_list.append('\n\n+==========')
    varstr_list.append(repr(type(val)) + ' ' + name + ' = ')
    varstr_list.append(six.text_type(val))
    if pad_stdout:
        varstr_list.append('L==========')
    varstr = '\n'.join(varstr_list)
    return varstr


def super_print(val, locals_=None):
    if locals_ is None:
        locals_ = get_parent_frame().f_locals
    print(get_varstr(val, locals_=locals_))


def print_keys(key_list, locals_=None):
    if locals_ is None:
        locals_ = get_parent_frame().f_locals
    strlist_ = parse_locals_keylist(locals_, key_list)
    print('\n'.join(strlist_))


def parse_locals_keylist(locals_, key_list, strlist_=None, prefix=''):
    """ For each key in keylist, puts its value in locals into a stringlist

    Args:
        locals_ (?):
        key_list (list):
        strlist_ (list): (default = None)
        prefix (unicode): (default = u'')

    Returns:
        list: strlist_

    CommandLine:
        python -m utool.util_dbg --exec-parse_locals_keylist

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_dbg import *  # NOQA
        >>> import utool as ut
        >>> locals_ = {'foo': [1, 2, 3], 'bar': 'spam', 'eggs': 4, 'num': 5}
        >>> key_list = [(len, 'foo'), 'bar.lower.__name__', 'eggs', 'num', 'other']
        >>> strlist_ = None
        >>> prefix = u''
        >>> strlist_ = parse_locals_keylist(locals_, key_list, strlist_, prefix)
        >>> result = ('strlist_ = %s' % (ut.repr2(strlist_, nl=True),))
        >>> print(result)
        strlist_ = [
            ' len(foo) = 3',
            " bar.lower.__name__ = 'lower'",
            ' eggs = 4',
            ' num = 5',
            ' other = NameError (this likely due to a misformatted printex and is not related to the exception)',
        ]
    """
    from utool import util_str
    if strlist_ is None:
        strlist_ = []

    for key in key_list:
        try:
            if key is None:
                strlist_.append('')
            elif isinstance(key, tuple):
                # Given a tuple of information
                tup = key
                func, key_ = tup
                val = get_varval_from_locals(key_, locals_)
                funcvalstr = six.text_type(func(val))
                callname = util_str.get_callable_name(func)
                strlist_.append('%s %s(%s) = %s' % (prefix, callname, key_, funcvalstr))
            elif isinstance(key, six.string_types):
                # Try to infer print from variable name
                val = get_varval_from_locals(key, locals_)
                #valstr = util_str.truncate_str(repr(val), maxlen=200)
                valstr = util_str.truncate_str(util_str.repr2(val), maxlen=200)
                strlist_.append('%s %s = %s' % (prefix, key, valstr))
            else:
                # Try to infer print from variable value
                val = key
                typestr = repr(type(val))
                namestr = get_varname_from_locals(val, locals_)
                #valstr = util_str.truncate_str(repr(val), maxlen=200)
                valstr = util_str.truncate_str(util_str.repr2(val), maxlen=200)
                strlist_.append('%s %s %s = %s' % (prefix, typestr, namestr, valstr))
        except AssertionError as ex:
            strlist_.append(prefix + ' ' + six.text_type(ex) + ' (this likely due to a misformatted printex and is not related to the exception)')
    return strlist_


def get_reprs(*args, **kwargs):
    if 'locals_' in kwargs:
        locals_ = kwargs['locals_']
    else:
        locals_ = locals()
        locals_.update(get_parent_frame().f_locals)

    msg_list = []
    var_list = list(args) + kwargs.get('var_list', [])
    for key in var_list:
        var = locals_[key]
        msg = util_str.horiz_string(six.text_type(key) + ' = ', repr(var))
        msg_list.append(msg)

    reprs = '\n' + util_str.indent('\n##\n'.join(msg_list)) + '\n'
    return reprs


def printvar2(varstr, attr='', typepad=0):
    locals_ = get_parent_frame().f_locals
    printvar(locals_, varstr, attr, typepad=typepad)


def printvar(locals_, varname, attr='.shape', typepad=0):
    #npprintopts = np.get_printoptions()
    #np.set_printoptions(threshold=5)
    dotpos = varname.find('.')
    # Locate var
    if dotpos == -1:
        var = locals_[varname]
    else:
        varname_ = varname[:dotpos]
        dotname_ = varname[dotpos:]
        var_ = locals_[varname_]  # NOQA
        var = eval('var_' + dotname_)
    # Print in format
    typestr = six.text_type(util_type.get_type(var)).ljust(typepad)

    if isinstance(var, np.ndarray):
        varstr = eval('str(var' + attr + ')')
        print('[var] %s %s = %s' % (typestr, varname + attr, varstr))
    elif isinstance(var, list):
        if attr == '.shape':
            func = 'len'
        else:
            func = ''
        varstr = eval('str(' + func + '(var))')
        print('[var] %s len(%s) = %s' % (typestr, varname, varstr))
    else:
        print('[var] %s %s = %r' % (typestr, varname, var))
    #np.set_printoptions(**npprintopts)


#def my_numpy_printops(linewidth=200, threshold=500, precision=8, edgeitems=5):
#    np.set_printoptions(linewidth=linewidth,
#                        precision=precision,
#                        edgeitems=edgeitems,
#                        threshold=threshold)


class EmbedOnException(object):
    """
    Context manager which embeds in ipython if an exception is thrown
    """
    def __init__(self):
        pass

    def __enter__(self):
        return self

    def __call__(self):
        return self

    def __exit__(self, type_, value, trace):
        if trace is not None:
            print('!!! EMBED ON EXCEPTION !!!')
            print('[util_dbg] %r in context manager!: %s ' % (type_, str(value)))
            import utool as ut
            import traceback
            traceback.print_exception(type_, value, trace)
            # Grab the context of the frame where the failure occurred
            trace_globals = trace.tb_frame.f_globals
            trace_locals = trace.tb_frame.f_locals
            trace_ns = trace_globals.copy()
            trace_ns.update(trace_locals)
            # Hack to bring back self
            if 'self' in trace_ns:
                self = trace_ns['self']
            # execstr_trace_g = ut.execstr_dict(trace_globals, 'trace_globals')
            # execstr_trace_l = ut.execstr_dict(trace_locals, 'trace_locals')
            # execstr_trace = execstr_trace_g + '\n' + execstr_trace_l
            # exec(execstr_trace)
            locals().update(trace_ns)
            ut.embed()

# maybe this will be easier to type?
embed_on_exception_context = EmbedOnException()
eoxc = embed_on_exception_context

if __name__ == '__main__':
    """
    CommandLine:
        python -c "import utool, utool.util_dbg; utool.doctest_funcs(utool.util_dbg, allexamples=True)"
        python -c "import utool, utool.util_dbg; utool.doctest_funcs(utool.util_dbg)"
        python -m utool.util_dbg
        python -m utool.util_dbg --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
Checks python files in a project for consistent header patterns

Rectify with ensure_python3_compatible
"""
from __future__ import absolute_import, division, print_function, unicode_literals


def find_encoding_insert_position(line_list):
    linenum = None
    for linenum, line in enumerate(line_list):
        # Ensure encoding line is after shebang
        if line.startswith('#!'):
            continue
        break
    if len(line_list) == 0 or linenum >= len(line_list) - 1:
        print(' ! could not find approprate position')
        linenum = None
    return linenum


if __name__ == '__main__':
    """
    python -m utool.util_scripts.pyproj_checker
    python ~/code/utool/utool/util_scripts/pyproj_checker.py
    """
    import utool as ut
    import re
    exclude_dirs = ['_broken', '_doc', 'build']
    fpath_list = ut.glob('.', '*.py', exclude_dirs=exclude_dirs, recursive=True)

    encoding_line = '# -*- coding: utf-8 -*-'

    pattern_items = [
        encoding_line,
        #'from __future__ import absolute_import, division, print_function',
        #'from __future__ import absolute_import, division, print_function, unicode_literals',
    ]

    show_diff = ut.get_argflag('--diff')
    do_write = ut.get_argflag('--write')

    need_encoding_fpaths = []

    for pat in pattern_items:
        print('Checking for pattern: %r' % (pat,))
        for fpath in fpath_list:
            pattern = re.escape(pat)
            found_lines, found_lxs = ut.grepfile(fpath, pattern)
            # DID NOT FIND ENCODING LINE
            if len(found_lines) == 0:
                need_encoding_fpaths.append(fpath)

    print('The following fpaths need encoding lines: ' + ut.repr4(need_encoding_fpaths, strvals=True))

    if do_write or show_diff:
        for fpath in need_encoding_fpaths:
            print('\n-----------------\nFound file without encodeing line: ' + fpath)
            line_list = ut.read_lines_from(fpath)
            linenum = find_encoding_insert_position(line_list)
            if linenum is not None:
                #print(' * linenum = %r' % (linenum,))
                new_lines = line_list[:linenum] + [encoding_line + '\n'] + line_list[linenum:]
                new_text = ''.join(new_lines)
                if show_diff:
                    old_text = ''.join(line_list)
                    textdiff = ut.get_textdiff(old_text, new_text, num_context_lines=1)
                    print('Diff:')
                    ut.print_difftext(textdiff)
                if do_write:
                    ut.writeto(fpath, new_text)
                    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import six


def ensure_iterable(obj):
    r"""
    Args:
        obj (scalar or iterable):

    Returns:
        it3erable: obj if it was iterable otherwise [obj]

    CommandLine:
        python -m utool._internal.meta_util_iter --test-ensure_iterable

    Timeit:
        %timeit ut.ensure_iterable([1])
        %timeit ut.ensure_iterable(1)
        %timeit ut.ensure_iterable(np.array(1))
        %timeit ut.ensure_iterable([1])
        %timeit [1]


    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool._internal.meta_util_iter import *  # NOQA
        >>> # build test data
        >>> obj_list = [3, [3], '3', (3,), [3,4,5]]
        >>> # execute function
        >>> result = [ensure_iterable(obj) for obj in obj_list]
        >>> # verify results
        >>> result = str(result)
        >>> print(result)
        [[3], [3], ['3'], (3,), [3, 4, 5]]
    """
    if isiterable(obj):
        return obj
    else:
        return [obj]


def isiterable(obj):
    """
    Returns if the object can be iterated over and is NOT a string
    # TODO: implement isscalar similar to numpy

    Args:
        obj (scalar or iterable):

    Returns:
        bool:

    CommandLine:
        python -m utool._internal.meta_util_iter --test-isiterable

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool._internal.meta_util_iter import *  # NOQA
        >>> # build test data
        >>> obj_list = [3, [3], '3', (3,), [3,4,5]]
        >>> # execute function
        >>> result = [isiterable(obj) for obj in obj_list]
        >>> # verify results
        >>> print(result)
        [False, True, False, True, True]
    """
    try:
        iter(obj)
        return not isinstance(obj, six.string_types)
    except Exception:
        return False
    #return np.iterable(obj) and not isinstance(obj, six.string_types)


def isscalar(obj):
    return not isiterable(obj)


#def isscalar(obj):
if __name__ == '__main__':
    """
    CommandLine:
        python -m utool._internal.meta_util_iter
        python -m utool._internal.meta_util_iter --allexamples
        python -m utool._internal.meta_util_iter --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
from six.moves import builtins
import utool
print, rrr, profile = utool.inject2(__name__)


@utool.indent_func
def func1():
    print('enter func1')
    print('exit  func1')


@utool.indent_func
def func2():
    print('enter func2')
    func1()
    print('exit  func2')


def remove_timestamp(string):
    import re
    return re.sub(r'\[\d\d:\d\d:\d\d\]', '', string)


@utool.indent_func
def test():
    print('enter test')
    log_fpath1 = utool.get_app_resource_dir('utool', 'test_logfile1.txt')
    log_fpath2 = utool.get_app_resource_dir('utool', 'test_logfile2.txt')

    utool.start_logging(log_fpath1, 'w')
    func1()
    func2()
    utool.stop_logging()

    print('\n\n')
    print('This line is NOT logged')
    print('\n\n')

    utool.start_logging(log_fpath2, 'w')
    print('This line is logged')
    utool.stop_logging()

    log1 = utool.read_from(log_fpath1, verbose=False)
    log2 = utool.read_from(log_fpath2, verbose=False)

    target1 = utool.unindent('''
    <__LOG_START__>
    logging to log_fpath=%r
    [test][func1]enter func1
    [test][func1]exit  func1
    [test][func2]enter func2
    [test][func2][func1]enter func1
    [test][func2][func1]exit  func1
    [test][func2]exit  func2
    <__LOG_STOP__>''' % log_fpath1).strip()

    target2 = utool.unindent('''
    <__LOG_START__>
    logging to log_fpath=%r
    [test]This line is logged
    <__LOG_STOP__>''' % log_fpath2).strip()

    output1 = remove_timestamp(log1).strip()
    output2 = remove_timestamp(log2).strip()

    try:
        assert target1 == output1, 'target1 failed'
        assert target2 == output2, 'target2 failed'
        builtins.print('TEST PASSED')
    except AssertionError:
        builtins.print('\n<!!! TEST FAILED !!!>')

        builtins.print('\ntarget1:')
        builtins.print(target1)
        builtins.print('\noutput1:')
        builtins.print(output1)

        builtins.print('\ntarget2:')
        builtins.print(target2)
        builtins.print('\noutput2:')
        builtins.print(output2)

        builtins.print('</!!! TEST FAILED !!!>\n')
        raise


if __name__ == '__main__':
    test()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
from os.path import join, splitext, basename
from utool import util_inject
print, rrr, profile = util_inject.inject2(__name__)


def add_new_mimetype_association(ext, mime_name, exe_fpath=None, dry=True):
    """
    TODO: move to external manager and generalize

    Args:
        ext (str): extension to associate
        mime_name (str): the name of the mime_name to create (defaults to ext)
        exe_fpath (str): executable location if this is for one specific file

    References:
        https://wiki.archlinux.org/index.php/Default_applications#Custom_file_associations

    Args:
        ext (str): extension to associate
        exe_fpath (str): executable location
        mime_name (str): the name of the mime_name to create (defaults to ext)

    CommandLine:
        python -m utool.util_ubuntu --exec-add_new_mimetype_association
        # Add ability to open ipython notebooks via double click
        python -m utool.util_ubuntu --exec-add_new_mimetype_association --mime-name=ipynb+json --ext=.ipynb --exe-fpath=/usr/local/bin/ipynb
        python -m utool.util_ubuntu --exec-add_new_mimetype_association --mime-name=ipynb+json --ext=.ipynb --exe-fpath=jupyter-notebook --force

        python -m utool.util_ubuntu --exec-add_new_mimetype_association --mime-name=sqlite --ext=.sqlite --exe-fpath=sqlitebrowser

    Example:
        >>> # DISABLE_DOCTEST
        >>> # SCRIPT
        >>> from utool.util_ubuntu import *  # NOQA
        >>> import utool as ut
        >>> ext = ut.get_argval('--ext', type_=str, default=None)
        >>> mime_name = ut.get_argval('--mime_name', type_=str, default=None)
        >>> exe_fpath = ut.get_argval('--exe_fpath', type_=str, default=None)
        >>> dry = not ut.get_argflag('--force')
        >>> result = add_new_mimetype_association(ext, mime_name, exe_fpath, dry)
        >>> print(result)
    """
    import utool as ut
    terminal = True

    mime_codeblock = ut.codeblock(
        '''
        <?xml version="1.0" encoding="UTF-8"?>
        <mime-info xmlns="http://www.freedesktop.org/standards/shared-mime-info">
            <mime-type type="application/x-{mime_name}">
                <glob-deleteall/>
                <glob pattern="*{ext}"/>
            </mime-type>
        </mime-info>
        '''
    ).format(**locals())

    prefix = ut.truepath('~/.local/share')
    mime_dpath = join(prefix, 'mime/packages')
    mime_fpath = join(mime_dpath, 'application-x-{mime_name}.xml'.format(**locals()))

    print(mime_codeblock)
    print('---')
    print(mime_fpath)
    print('L___')

    if exe_fpath is not None:
        exe_fname_noext = splitext(basename(exe_fpath))[0]
        app_name = exe_fname_noext.replace('_', '-')
        nice_name = ' '.join(
            [word[0].upper() + word[1:].lower()
             for word in app_name.replace('-', ' ').split(' ')]
        )
        app_codeblock = ut.codeblock(
            '''
            [Desktop Entry]
            Name={nice_name}
            Exec={exe_fpath}
            MimeType=application/x-{mime_name}
            Terminal={terminal}
            Type=Application
            Categories=Utility;Application;
            Comment=Custom App
            '''
        ).format(**locals())
        app_dpath = join(prefix, 'applications')
        app_fpath = join(app_dpath, '{app_name}.desktop'.format(**locals()))

        print(app_codeblock)
        print('---')
        print(app_fpath)
        print('L___')

    # WRITE FILES
    if not dry:
        ut.ensuredir(mime_dpath)
        ut.ensuredir(app_dpath)

        ut.writeto(mime_fpath, mime_codeblock, verbose=ut.NOT_QUIET, n=None)
        if exe_fpath is not None:
            ut.writeto(app_fpath, app_codeblock, verbose=ut.NOT_QUIET, n=None)

        # UPDATE BACKENDS

        #ut.cmd('update-mime-database /usr/share/mime')
        #~/.local/share/applications/mimeapps.list
        print(ut.codeblock(
            '''
            Run these commands:
            update-desktop-database ~/.local/share/applications
            update-mime-database ~/.local/share/mime
            '''
        ))
        if exe_fpath is not None:
            ut.cmd('update-desktop-database ~/.local/share/applications')
        ut.cmd('update-mime-database ~/.local/share/mime')
    else:
        print('dry_run')


def make_application_icon(exe_fpath, dry=True, props={}):
    r"""
    CommandLine:
        python -m utool.util_ubuntu --exec-make_application_icon --exe=cockatrice --icon=/home/joncrall/code/Cockatrice/cockatrice/resources/cockatrice.png
        python -m utool.util_ubuntu --exec-make_application_icon --exe=cockatrice --icon=/home/joncrall/code/Cockatrice/cockatrice/resources/cockatrice.png
        python -m utool.util_ubuntu --exec-make_application_icon --exe=/opt/zotero/zotero --icon=/opt/zotero/chrome/icons/default/main-window.ico

        python -m utool.util_ubuntu --exec-make_application_icon --exe "env WINEPREFIX="/home/joncrall/.wine" wine C:\\\\windows\\\\command\\\\start.exe /Unix /home/joncrall/.wine32-dotnet45/dosdevices/c:/users/Public/Desktop/Hearthstone.lnk" --path "/home/joncrall/.wine/dosdevices/c:/Program Files (x86)/Hearthstone"
        # Exec=env WINEPREFIX="/home/joncrall/.wine" wine /home/joncrall/.wine/drive_c/Program\ Files\ \(x86\)/Battle.net/Battle.net.exe

        --icon=/opt/zotero/chrome/icons/default/main-window.ico

        python -m utool.util_ubuntu --exec-make_application_icon --exe=/home/joncrall/code/build-ArenaTracker-Desktop_Qt_5_6_1_GCC_64bit-Debug

        update-desktop-database ~/.local/share/applications


    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_ubuntu import *  # NOQA
        >>> import utool as ut
        >>> exe_fpath = ut.get_argval('--exe', default='cockatrice')
        >>> icon = ut.get_argval('--icon', default=None)
        >>> dry = not ut.get_argflag(('--write', '-w'))
        >>> props = {'terminal': False, 'icon': icon}
        >>> result = make_application_icon(exe_fpath, dry, props)
        >>> print(result)
    """
    import utool as ut
    exe_fname_noext = splitext(basename(exe_fpath))[0]
    app_name = exe_fname_noext.replace('_', '-')
    nice_name = ' '.join(
        [word[0].upper() + word[1:].lower()
         for word in app_name.replace('-', ' ').split(' ')]
    )
    lines = [
        '[Desktop Entry]',
        'Name={nice_name}',
        'Exec={exe_fpath}',
    ]

    if 'mime_name' in props:
        lines += ['MimeType=application/x-{mime_name}']

    if 'icon' in props:
        lines += ['Icon={icon}']

    if props.get('path'):
        lines += ['Path={path}']

    # if props.get('comment'):
    #     lines += ['Path={comment}']

    lines += [
        'Terminal={terminal}',
        'Type=Application',
        'Categories=Utility;Application;',
        'Comment=Custom App',
    ]
    fmtdict = locals()
    fmtdict.update(props)

    prefix = ut.truepath('~/.local/share')
    app_codeblock = '\n'.join(lines).format(**fmtdict)
    app_dpath = join(prefix, 'applications')
    app_fpath = join(app_dpath, '{app_name}.desktop'.format(**locals()))

    print(app_codeblock)
    print('---')
    print(app_fpath)
    print('L___')

    if not dry:
        ut.writeto(app_fpath, app_codeblock, verbose=ut.NOT_QUIET, n=None)
        ut.cmd('update-desktop-database ~/.local/share/applications')


class XCtrl(object):
    r"""
    xdotool key ctrl+shift+i

    References:
        http://superuser.com/questions/382616/detecting-currently-active-window
        http://askubuntu.com/questions/455762/xbindkeys-wont-work-properly

    Ignore:
        xdotool keyup --window 0 7 type --clearmodifiers ---window 0 '%paste'

        # List current windows:
        wmctrl  -l

        # Get current window
        xdotool getwindowfocus getwindowname


        #====
        # Get last opened window
        #====

        win_title=x-terminal-emulator.X-terminal-emulator
        key_ = 'x-terminal-emulator.X-terminal-emulator'

        # Get all windows in current workspace
        workspace_number=`wmctrl -d | grep '\*' | cut -d' ' -f 1`
        win_list=`wmctrl -lx | grep $win_title | grep " $workspace_number " | awk '{print $1}'`

        # Get stacking order of windows in current workspace
        win_order=$(xprop -root|grep "^_NET_CLIENT_LIST_STACKING" | tr "," " ")
        echo $win_order

    CommandLine:
        python -m utool.util_ubuntu XCtrl

    Example:
        >>> # DISABLE_DOCTEST
        >>> # Script
        >>> import utool as ut
        >>> from utool import util_ubuntu
        >>> orig_window = []
        >>> ut.copy_text_to_clipboard(ut.lorium_ipsum())
        >>> doscript = [
        >>>     ('focus', 'x-terminal-emulator.X-terminal-emulator'),
        >>>     ('type', '%paste'),
        >>>     ('key', 'KP_Enter'),
        >>>    # ('focus', 'GVIM')
        >>> ]
        >>> util_ubuntu.XCtrl.do(*doscript, sleeptime=.01)

    Ignore:
        >>> ut.copy_text_to_clipboard(text)
        >>> if '\n' in text or len(text) > 20:
        >>>     text = '\'%paste\''
        >>> else:
        >>>     import pipes
        >>>     text = pipes.quote(text.lstrip(' '))
        >>>     ('focus', 'GVIM'),
        >>> #
        >>> doscript = [
        >>>     ('focus', 'x-terminal-emulator.X-terminal-emulator'),
        >>>     ('type', text),
        >>>     ('key', 'KP_Enter'),
        >>> ]
        >>> ut.util_ubuntu.XCtrl.do(*doscript, sleeptime=.01)


    """
    # @staticmethod
    # def send_raw_key_input(keys):
    #     import utool as ut
    #     print('send key input: %r' % (keys,))
    #     args = ['xdotool', 'type', keys]
    #     ut.cmd(*args, quiet=True, silence=True)

    @staticmethod
    def move_window(win_key, bbox):
        """
        CommandLine:
            # List windows
            wmctrl -l
            # List desktops
            wmctrl -d

            # Window info
            xwininfo -id 60817412

            python -m utool.util_ubuntu XCtrl.move_window joncrall 0+1920,680,400,600,400
            python -m utool.util_ubuntu XCtrl.move_window joncrall [0,0,1000,1000]
            python -m utool.util_ubuntu XCtrl.move_window GVIM special2
            python -m utool.util_ubuntu XCtrl.move_window joncrall special2
            python -m utool.util_ubuntu XCtrl.move_window x-terminal-emulator.X-terminal-emulator [0,0,1000,1000]

        # >>> import utool as ut
        # >>> from utool import util_ubuntu
        # >>> orig_window = []
        # >>> X = util_ubuntu.XCtrl
        win_key =  'x-terminal-emulator.X-terminal-emulator'
        win_id = X.findall_window_ids(key)[0]

        python -m utool.util_ubuntu XCtrl.findall_window_ids gvim --src

        """
        import utool as ut
        try:
            import plottool_ibeis.screeninfo as screeninfo
        except ImportError:
            import plottool.screeninfo as screeninfo
        monitor_infos = {
            i + 1: screeninfo.get_resolution_info(i)
            for i in range(2)
        }
        # TODO: cut out borders
        # TODO: fix screeninfo monitor offsets
        # TODO: dynamic num screens
        def rel_to_abs_bbox(m, x, y, w, h):
            """ monitor_num, relative x, y, w, h """
            minfo = monitor_infos[m]
            # print('minfo(%d) = %s' % (m, ut.repr3(minfo),))
            mx, my = minfo['off_x'], minfo['off_y']
            mw, mh = minfo['pixels_w'], minfo['pixels_h']
            # Transform to the absolution position
            abs_x = (x * mw) + mx
            abs_y = (y * mh) + my
            abs_w = (w * mw)
            abs_h = (h * mh)
            abs_bbox = [abs_x, abs_y, abs_w, abs_h]
            abs_bbox = ','.join(map(str, map(int, abs_bbox)))
            return abs_bbox

        if win_key.startswith('joncrall') and bbox == 'special2':
            # Specify the relative position
            abs_bbox = rel_to_abs_bbox(m=2,
                                       x=0.0, y=0.7,
                                       w=1.0, h=0.3)
        elif win_key.startswith('GVIM') and bbox == 'special2':
            # Specify the relative position
            abs_bbox = rel_to_abs_bbox(m=2,
                                       x=0.0, y=0.0,
                                       w=1.0, h=0.7)
        else:
            abs_bbox = ','.join(map(str, eval(bbox)))

        print('MOVING: win_key = %r' % (win_key,))
        print('TO: abs_bbox = %r' % (abs_bbox,))
        # abs_bbox.replace('[', '').replace(']', '')
        # get = lambda cmd: ut.cmd2(' '.join(["/bin/bash", "-c", cmd]))['out']  # NOQA
        win_id = XCtrl.find_window_id(win_key, error='raise')
        print('MOVING: win_id = %r' % (win_id,))
        fmtdict = locals()
        cmd_list = [
            ("wmctrl -ir {win_id} -b remove,maximized_horz".format(**fmtdict)),
            ("wmctrl -ir {win_id} -b remove,maximized_vert".format(**fmtdict)),
            ("wmctrl -ir {win_id} -e 0,{abs_bbox}".format(**fmtdict)),
        ]
        print('\n'.join(cmd_list))
        for cmd in cmd_list:
            ut.cmd2(cmd)

    @staticmethod
    def findall_window_ids(pattern):
        """
        CommandLine:
            wmctrl  -l
            python -m utool.util_ubuntu XCtrl.findall_window_ids gvim --src
            python -m utool.util_ubuntu XCtrl.findall_window_ids gvim --src
            python -m utool.util_ubuntu XCtrl.findall_window_ids joncrall --src

        xprop -id

        wmctrl -l | awk '{print $1}' | xprop -id

        0x00a00007 | grep "WM_CLASS(STRING)"

        """
        import utool as ut
        cmdkw = dict(verbose=False, quiet=True, silence=True)
        command = "wmctrl -lx | grep '%s' | awk '{print $1}'" % (pattern,)
        # print(command)
        winid_list = ut.cmd(command, **cmdkw)[0].strip().split('\n')
        winid_list = [h for h in winid_list if h]
        winid_list = [int(h, 16) for h in winid_list]
        return winid_list

    @staticmethod
    def sort_window_ids(winid_list, order='mru'):
        """
        Orders window ids by most recently used
        """
        import utool as ut
        winid_order = XCtrl.sorted_window_ids(order)
        sorted_win_ids = ut.isect(winid_order, winid_list)
        return sorted_win_ids

    @staticmethod
    def killold(pattern, num=4):
        """
        Leaves no more than `num` instances of a program alive.  Ordering is
        determined by most recent usage.

        CommandLine:
            python -m utool.util_ubuntu XCtrl.killold gvim 2

        >>> import utool as ut
        >>> from utool import util_ubuntu
        >>> XCtrl = util_ubuntu.XCtrl
        >>> pattern = 'gvim'
        >>> num = 2

        """
        import utool as ut
        cmdkw = dict(verbose=False, quiet=True, silence=True)
        num = int(num)
        winid_list = XCtrl.findall_window_ids(pattern)
        winid_list = XCtrl.sort_window_ids(winid_list, 'mru')[num:]
        output_lines = ut.cmd(
            """wmctrl -lxp | awk '{print $1 " " $3}'""",
            **cmdkw)[0].strip().split('\n')
        output_fields = [line.split(' ') for line in output_lines]
        output_fields = [(int(wid, 16), int(pid)) for wid, pid in output_fields]
        pid_list = [pid for wid, pid in output_fields if wid in winid_list]
        import psutil
        for pid in pid_list:
            proc = psutil.Process(pid=pid)
            proc.kill()

    @staticmethod
    def sorted_window_ids(order='mru'):
        """
        Returns window ids orderd by criteria
        default is mru (most recently used)

        CommandLine:
            xprop -root | grep "^_NET_CLIENT_LIST_STACKING" | tr "," " "
            python -m utool.util_ubuntu XCtrl.sorted_window_ids
        """
        import utool as ut
        if order in ['mru', 'lru']:
            cmdkw = dict(verbose=False, quiet=True, silence=True)
            winid_order_str = ut.cmd(
                'xprop -root | grep "^_NET_CLIENT_LIST_STACKING"', **cmdkw)[0]
            winid_order = winid_order_str.split('#')[1].strip().split(', ')[::-1]
            winid_order = [int(h, 16) for h in winid_order]
            if order == 'lru':
                winid_order = winid_order[::-1]
        else:
            raise NotImplementedError(order)
        return winid_order

    @staticmethod
    def find_window_id(pattern, method='mru', error='raise'):
        """
        xprop -id 0x00a00007 | grep "WM_CLASS(STRING)"
        """
        import utool as ut
        winid_candidates = XCtrl.findall_window_ids(pattern)
        if len(winid_candidates) == 0:
            if error == 'raise':
                available_windows = ut.cmd2('wmctrl -l')['out']
                msg = 'No window matches pattern=%r' % (pattern,)
                msg += '\navailable windows are:\n%s' % (available_windows,)
                print(msg)
                raise Exception(msg)
            win_id = None
        elif len(winid_candidates) == 1:
            win_id = winid_candidates[0]
        else:
            # print('Multiple (%d) windows matches pattern=%r' % (
            #     len(winid_list), pattern,))
            # Find most recently used window with the focus name.
            win_id = XCtrl.sort_window_ids(winid_candidates, method)[0]
        return win_id

    @staticmethod
    def current_gvim_edit(op='e', fpath=''):
        r"""
        CommandLine:
            python -m utool.util_ubuntu XCtrl.current_gvim_edit sp ~/.bashrc
        """
        import utool as ut
        fpath = ut.unexpanduser(ut.truepath(fpath))
        # print('fpath = %r' % (fpath,))
        ut.copy_text_to_clipboard(fpath)
        # print(ut.get_clipboard())
        doscript = [
            ('focus', 'gvim'),
            ('key', 'Escape'),
            ('type2', ';' + op + ' ' + fpath),
            # ('type2', ';' + op + ' '),
            # ('key', 'ctrl+v'),
            ('key', 'KP_Enter'),
        ]
        XCtrl.do(*doscript, verbose=0, sleeptime=.001)

    @staticmethod
    def copy_gvim_to_terminal_script(text, return_to_win="1", verbose=0, sleeptime=.02):
        """
        import utool.util_ubuntu
        utool.util_ubuntu.XCtrl.copy_gvim_to_terminal_script('print("hi")', verbose=1)
        python -m utool.util_ubuntu XCtrl.copy_gvim_to_terminal_script "echo hi" 1 1

        If this doesn't work make sure pyperclip is installed and set to xsel

        print('foobar')
        echo hi
        """
        # Prepare to send text to xdotool
        import utool as ut
        import utool.util_ubuntu
        ut.copy_text_to_clipboard(text)

        if verbose:
            print('text = %r' % (text,))
            print(ut.get_clipboard())

        import re
        terminal_pattern = r'\|'.join([
            'terminal',
            re.escape('terminator.Terminator'),  # gtk3 terminator
            re.escape('x-terminal-emulator.X-terminal-emulator'),  # gtk2 terminator
        ])

        # Build xdtool script
        doscript = [
            ('remember_window_id', 'ACTIVE_WIN'),
            # ('focus', 'x-terminal-emulator.X-terminal-emulator'),
            ('focus', terminal_pattern),
            ('key', 'ctrl+shift+v'),
            ('key', 'KP_Enter'),
        ]
        if '\n' in text:
            # Press enter twice for multiline texts
            doscript += [
                ('key', 'KP_Enter'),
            ]

        if return_to_win == "1":
            doscript += [
                ('focus_id', '$ACTIVE_WIN'),
            ]
        # execute script
        # verbose = 1
        utool.util_ubuntu.XCtrl.do(*doscript, sleeptime=sleeptime, verbose=verbose)

    @staticmethod
    def do(*cmd_list, **kwargs):
        import utool as ut
        import time
        import six
        import sys
        verbose = kwargs.get('verbose', False)
        orig_print = globals()['print']
        print = ut.partial(orig_print, file=kwargs.get('file', sys.stdout))
        # print('Running xctrl.do script')
        if verbose:
            print('Executing x do: %s' % (ut.repr4(cmd_list),))
        debug = False

        cmdkw = dict(verbose=False, quiet=True, silence=True)
        # http://askubuntu.com/questions/455762/xbindkeys-wont-work-properly
        # Make things work even if other keys are pressed
        defaultsleep = 0.0
        sleeptime = kwargs.get('sleeptime', defaultsleep)
        time.sleep(.05)
        out, err, ret = ut.cmd('xset r off', **cmdkw)
        if debug:
            print('----------')
            print('xset r off')
            print('ret = %r' % (ret,))
            print('err = %r' % (err,))
            print('out = %r' % (out,))

        memory = {}

        tmpverbose = 0
        for count, item in enumerate(cmd_list):
            # print('item = %r' % (item,))
            sleeptime = kwargs.get('sleeptime', defaultsleep)
            if tmpverbose:
                print('moving on')
            tmpverbose = 0
            nocommand = 0

            assert isinstance(item, tuple)
            assert len(item) >= 2
            xcmd, key_ = item[0:2]
            if len(item) >= 3:
                if isinstance(item[2], six.string_types) and item[2].endswith('?'):
                    sleeptime = float(item[2][:-1])
                    tmpverbose = 1
                    print('special command sleep')
                    print('sleeptime = %r' % (sleeptime,))
                else:
                    sleeptime = float(item[2])

            if xcmd == 'focus':
                key_ = str(key_)
                if key_.startswith('$'):
                    key_ = memory[key_[1:]]
                pattern = key_
                win_id = XCtrl.find_window_id(pattern, method='mru')
                if win_id is None:
                    args = ['wmctrl', '-xa', pattern]
                else:
                    args = ['wmctrl', '-ia', hex(win_id)]
            elif xcmd == 'focus_id':
                key_ = str(key_)
                if key_.startswith('$'):
                    key_ = memory[key_[1:]]
                args = ['wmctrl', '-ia', hex(key_)]
            elif xcmd == 'remember_window_id':
                out, err, ret = ut.cmd('xdotool getwindowfocus', **cmdkw)
                memory[key_] = int(out.strip())
                nocommand = True
                args = []
            elif xcmd == 'remember_window_name':
                out, err, ret = ut.cmd('xdotool getwindowfocus getwindowname', **cmdkw)
                import pipes
                memory[key_] = pipes.quote(out.strip())
                nocommand = True
                args = []
            elif xcmd == 'type':
                args = [
                    'xdotool',
                    'keyup', '--window', '0', '7',
                    'type', '--clearmodifiers',
                    '--window', '0', str(key_)
                ]
            elif xcmd == 'type2':
                import pipes
                args = [
                    'xdotool', 'type', pipes.quote(str(key_))
                ]
            elif xcmd == 'xset-r-on':
                args = ['xset', 'r', 'on']
            elif xcmd == 'xset-r-off':
                args = ['xset', 'r', 'off']
            else:
                args = ['xdotool', str(xcmd), str(key_)]

            if verbose or tmpverbose:
                print('\n\n# Step %d' % (count,))
                print(args, ' '.join(args))

            if nocommand:
                continue
                # print('args = %r -> %s' % (args, ' '.join(args),))
            # print('args = %r' % (args,))
            out, err, ret = ut.cmd(*args, **cmdkw)
            if debug:
                print('---- ' + xcmd + ' ------')
                print(' '.join(args))
                print('ret = %r' % (ret,))
                print('err = %r' % (err,))
                print('out = %r' % (out,))

            if sleeptime > 0:
                time.sleep(sleeptime)

        out, err, ret = ut.cmd('xset r on', verbose=False, quiet=True,
                               silence=True)
        if debug:
            print('----------')
            print('xset r on')
            print('ret = %r' % (ret,))
            print('err = %r' % (err,))
            print('out = %r' % (out,))

    @staticmethod
    def focus_window(winhandle, path=None, name=None, sleeptime=.01):
        """
        sudo apt-get install xautomation
        apt-get install autokey-gtk

        wmctrl -xa gnome-terminal.Gnome-terminal
        wmctrl -xl
        """
        import utool as ut
        import time
        print('focus: ' + winhandle)
        args = ['wmctrl', '-xa', winhandle]
        ut.cmd(*args, verbose=False, quiet=True)
        time.sleep(sleeptime)


xctrl = XCtrl


def monitor_mouse():
    """
    CommandLine:
        python -m utool.util_ubuntu monitor_mouse

    Example:
        >>> # DISABLE_DOCTEST
        >>> # SCRIPT
        >>> from utool.util_ubuntu import *  # NOQA
        >>> import utool as ut
        >>> monitor_mouse()
    """
    import utool as ut
    import re
    import parse
    mouse_ids = ut.cmd('xinput --list ', verbose=False, quiet=True)[0]
    x = mouse_ids.decode('utf-8')
    pattern = 'mouse'
    pattern = 'trackball'
    print(x)
    grepres = ut.greplines(x.split('\n'), pattern, reflags=re.IGNORECASE)
    mouse_id = parse.parse('{left}id={id}{right}', grepres[0][0])['id']
    print('mouse_id = %r' % (mouse_id,))
    import time
    while True:
        time.sleep(.2)
        out = ut.cmd('xinput --query-state ' + mouse_id, verbose=False, quiet=True)[0]
        print(out)


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_ubuntu
        python -m utool.util_ubuntu --allexamples
        python -m utool.util_ubuntu --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
Utilities for IPython/Jupyter Notebooks


CommandLine:
    # to connect to a notebook on a remote machine that does not have the
    # appropriate port exposed you must start an SSH tunnel.
    # Typically a jupyter-notebook runs on port 8888.
    # Run this command on your local machine.
    ssh -N -f -L localhost:8888:localhost:8889 <remote_user>@<remote_host>
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import sys
from utool import util_inject
from collections import namedtuple
print, rrr, profile = util_inject.inject2(__name__)


IPYNBCell = namedtuple('IPYNBCell', ['header', 'code', 'footer'])


def make_autogen_str():
    r"""
    Returns:
        str:

    CommandLine:
        python -m utool.util_ipynb --exec-make_autogen_str --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_ipynb import *  # NOQA
        >>> import utool as ut
        >>> result = make_autogen_str()
        >>> print(result)
    """
    import utool as ut
    def get_regen_cmd():
        try:
            if len(sys.argv) > 0 and ut.checkpath(sys.argv[0]):
                # Check if running python command
                if ut.is_python_module(sys.argv[0]):
                    python_exe = ut.python_executable(check=False)
                    modname = ut.get_modname_from_modpath(sys.argv[0])
                    new_argv = [python_exe, '-m', modname] + sys.argv[1:]
                    return ' '.join(new_argv)
        except Exception as ex:
            ut.printex(ex, iswarning=True)
        return ' '.join(sys.argv)

    autogenkw = dict(
        stamp=ut.timestamp('printable'),
        regen_cmd=get_regen_cmd()
        #' '.join(sys.argv)
    )
    return ut.codeblock(
        '''
        # Autogenerated on {stamp}
        # Regen Command:
        #    {regen_cmd}
        #
        '''
    ).format(**autogenkw)


def export_notebook(run_nb, fname):
    import IPython.nbconvert.exporters
    import codecs
    #exporter = IPython.nbconvert.exporters.PDFExporter()
    exporter = IPython.nbconvert.exporters.HTMLExporter()
    output, resources = exporter.from_notebook_node(run_nb)
    ext = resources['output_extension']
    output_fpath = fname + ext
    #codecs.open(output_fname, 'w', encoding='utf-8').write(output)
    codecs.open(output_fpath, 'w').write(output)
    return output_fpath
    #IPython.nbconvert.exporters.export_python(runner.nb)


def run_ipython_notebook(notebook_str):
    """
    References:
        https://github.com/paulgb/runipy
        >>> from utool.util_ipynb import *  # NOQA
    """
    from runipy.notebook_runner import NotebookRunner
    import nbformat
    import logging
    log_format = '%(asctime)s %(levelname)s: %(message)s'
    log_datefmt = '%m/%d/%Y %I:%M:%S %p'
    logging.basicConfig(
        level=logging.INFO, format=log_format, datefmt=log_datefmt
    )
    #fpath = 'tmp.ipynb'
    #notebook_str = ut.readfrom(fpath)
    #nb3 = IPython.nbformat.reads(notebook_str, 3)
    #cell = nb4.cells[1]
    #self = runner
    #runner = NotebookRunner(nb3, mpl_inline=True)
    print('Executing IPython notebook')
    nb4 = nbformat.reads(notebook_str, 4)
    runner = NotebookRunner(nb4)
    runner.run_notebook(skip_exceptions=False)
    run_nb = runner.nb
    return run_nb


def normalize_cells(block):
    if isinstance(block, (tuple, list)):
        if len(block) == 2:
            header, code = block
            footer = None
        elif len(block) == 3:
            header, code, footer = block
        else:
            assert False
    else:
        header = None
        footer = None
        code = block
    return IPYNBCell(header, code, footer)


def format_cells(block, locals_=None):
    import utool as ut
    try:
        if isinstance(block, (tuple, list)):
            if len(block) == 2:
                header, code = block
                footer = None
            elif len(block) == 3:
                header, code, footer = block
            else:
                assert False
        else:
            header = None
            footer = None
            code = block

        if header is not None:
            replace_chars = [('’', '\''), ('–', '-')]  # NOQA
            for tup in replace_chars:
                header = header.replace(*tup)

        if footer is not None:
            replace_chars = [('’', '\''), ('–', '-')]  # NOQA
            for tup in replace_chars:
                footer = footer.replace(*tup)

        if locals_ is not None and code is not None:

            def modify_code_indent_formatdict(code, locals_):
                # Parse out search and replace locations in code
                ncl1 = ut.negative_lookbehind('{')
                ncl2 = ut.negative_lookahead('{')
                ncr1 = ut.negative_lookbehind('}')
                ncr2 = ut.negative_lookahead('}')
                left = ncl1 + '{' + ncl2
                right = ncr1 + '}' + ncr2
                fmtpat = left + ut.named_field('key', '[^}]*') + right
                spacepat = ut.named_field('indent', r'^\s+')
                pattern = spacepat + fmtpat
                import re
                seen_ = set([])
                for m in re.finditer(pattern, code, flags=re.MULTILINE):
                    indent = (m.groupdict()['indent'])
                    key = (m.groupdict()['key'])
                    if key in locals_ and key not in seen_:
                        seen_.add(key)
                        locals_[key] = ut.indent_rest(locals_[key], indent)

            modify_code_indent_formatdict(code, locals_)
            code_ = code.format(**locals_)
            ut.is_valid_python(code_, reraise=True, ipy_magic_workaround=True)
        else:
            code_ = code
        cell_list = []
        if header is not None:
            cell_list += [markdown_cell(header)]
        if code is not None:
            cell_list += [code_cell(code_)]
        if footer is not None:
            cell_list += [markdown_cell(footer)]
        return cell_list
    except Exception as ex:
        ut.printex(ex, 'failed to format cell', keys=['header', 'block'])
        raise


def code_cell(sourcecode):
    r"""
    Args:
        sourcecode (str):

    Returns:
        str: json formatted ipython notebook code cell

    CommandLine:
        python -m ibeis.templates.generate_notebook --exec-code_cell

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.templates.generate_notebook import *  # NOQA
        >>> sourcecode = notebook_cells.timestamp_distribution[1]
        >>> sourcecode = notebook_cells.initialize[1]
        >>> result = code_cell(sourcecode)
        >>> print(result)
    """
    import utool as ut
    sourcecode = ut.remove_codeblock_syntax_sentinals(sourcecode)
    cell_header = ut.codeblock(
        '''
        {
         "cell_type": "code",
         "execution_count": null,
         "metadata": {
          "collapsed": true
         },
         "outputs": [],
         "source":
        ''')
    cell_footer = ut.codeblock(
        '''
        }
        ''')
    if sourcecode is None:
        source_line_repr = ' []\n'
    else:
        lines = sourcecode.split('\n')
        line_list = [line + '\n' if count < len(lines) else line
                     for count, line in enumerate(lines, start=1)]
        #repr_line_list = [repr_single_for_md(line) for line in line_list]
        repr_line_list = [repr_single_for_md(line) for line in line_list]
        source_line_repr = ut.indent(',\n'.join(repr_line_list), ' ' * 2)
        source_line_repr = ' [\n' + source_line_repr + '\n ]\n'
    return (cell_header + source_line_repr + cell_footer)


def markdown_cell(markdown):
    r"""
    Args:
        markdown (str):

    Returns:
        str: json formatted ipython notebook markdown cell

    CommandLine:
        python -m ibeis.templates.generate_notebook --exec-markdown_cell

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.templates.generate_notebook import *  # NOQA
        >>> markdown = '# Title'
        >>> result = markdown_cell(markdown)
        >>> print(result)
    """
    import utool as ut
    markdown_header = ut.codeblock(
        '''
          {
           "cell_type": "markdown",
           "metadata": {},
           "source": [
        '''
    )
    markdown_footer = ut.codeblock(
        '''
           ]
          }
        '''
    )
    return (markdown_header + '\n' +
            ut.indent(repr_single_for_md(markdown), ' ' * 2) +
            '\n' + markdown_footer)


def make_notebook(cell_list):
    """
    References:
        # Change cell width
        http://stackoverflow.com/questions/21971449/how-do-i-increase-the-cell-width-of-the-ipython-notebook-in-my-browser/24207353#24207353
    """
    import utool as ut
    header = ut.codeblock(
        '''
        {
         "cells": [
        '''
    )

    footer = ut.codeblock(
        '''
         ],
         "metadata": {
          "kernelspec": {
           "display_name": "Python 2",
           "language": "python",
           "name": "python2"
          },
          "language_info": {
           "codemirror_mode": {
            "name": "ipython",
            "version": 2
           },
           "file_extension": ".py",
           "mimetype": "text/x-python",
           "name": "python",
           "nbconvert_exporter": "python",
           "pygments_lexer": "ipython2",
           "version": "2.7.6"
          }
         },
         "nbformat": 4,
         "nbformat_minor": 0
        }
        ''')

    cell_body = ut.indent(',\n'.join(cell_list), '  ')
    notebook_str = header + '\n' + cell_body +  '\n' +  footer
    try:
        import json
        json.loads(notebook_str)
    except ValueError as ex:
        ut.printex(ex, 'Invalid notebook JSON')
        raise
    return notebook_str


def repr_single_for_md(s):
    r"""
    Args:
        s (str):

    Returns:
        str: str_repr

    CommandLine:
        python -m ibeis.templates.generate_notebook --exec-repr_single_for_md --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from ibeis.templates.generate_notebook import *  # NOQA
        >>> s = '#HTML(\'<iframe src="%s" width=700 height=350></iframe>\' % pdf_fpath)'
        >>> result = repr_single_for_md(s)
        >>> print(result)
    """
    import utool as ut
    if True:
        str_repr = ut.reprfunc(s)
        import re
        if str_repr.startswith('\''):
            dq = (ut.DOUBLE_QUOTE)
            sq = (ut.SINGLE_QUOTE)
            bs = (ut.BACKSLASH)
            dq_, sq_, bs_ = list(map(re.escape, [dq, sq, bs]))
            no_bs = ut.negative_lookbehind(bs_)
            #no_sq = ut.negative_lookbehind(sq)
            #no_dq = ut.negative_lookbehind(dq)

            #inside = str_repr[1:-1]
            #inside = re.sub(no_bs + dq, bs + dq, inside)
            #inside = re.sub(no_bs + bs + sq, r"\\'", r"'", inside)
            #str_repr = '"' + inside + '"'
            #inside = re.sub(r'"', r'\\"', inside)
            #inside = re.sub(ut.negative_lookbehind(r"'") + r"\\'", r"'", inside)

            inside = str_repr[1:-1]
            # Escape double quotes
            inside = re.sub(no_bs + r'"', r'\\"', inside)
            # Unescape single quotes
            inside = re.sub(no_bs + bs_ + r"'", r"'", inside)
            # Append external double quotes
            str_repr = '"' + inside + '"'
        return str_repr
    else:
        return '"' + ut.reprfunc('\'' + s)[2:]


if __name__ == '__main__':
    r"""
    CommandLine:
        python -m utool.util_ipynb
        python -m utool.util_ipynb --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()
